"""Utility functions for PyMC"""


# License: Scipy compatible
# Author: David Huard, 2006

import numpy as np
import sys
import select
import os
import time
from copy import copy
from .PyMCObjects import Variable
from . import flib
import pdb
from numpy.linalg.linalg import LinAlgError
from numpy.linalg import cholesky, eigh
from .Node import logp_of_set, logp_gradient_of_set
import types
from .datatypes import *
import inspect

from . import six
from .six import print_
from functools import reduce
reduce = six.moves.reduce

from numpy import (sqrt, ndarray, asmatrix, array, prod,
                   asarray, atleast_1d, iterable, linspace, diff,
                   around, log10, zeros, arange, digitize, apply_along_axis,
                   concatenate, bincount, sort, hsplit, argsort, inf, shape,
                   ndim, swapaxes, ravel, diag, cov, transpose as tr)

__all__ = ['append', 'check_list', 'autocorr', 'calc_min_interval',
           'check_type', 'ar1',
           'ar1_gen', 'draw_random', 'histogram', 'hpd', 'invcdf',
           'make_indices', 'normcdf', 'quantiles', 'rec_getattr',
           'rec_setattr', 'round_array', 'trace_generator', 'msqrt', 'safe_len',
           'log_difference', 'find_generations', 'crawl_dataless', 'logit',
           'invlogit', 'stukel_logit', 'stukel_invlogit', 'symmetrize', 'value',
           'get_signature']

symmetrize = flib.symmetrize

PY3 = sys.version_info >= (3, 3)

def get_signature_py3(func):
    (args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults,
        annotations) = inspect.getfullargspec(func)
    return args, defaults
    
def get_signature_py2(func):
    (args, varargs, varkw, defaults) = inspect.getargspec(func)
    return args, defaults
    
    
if PY3:
    get_signature = get_signature_py3
else:
    get_signature = get_signature_py2
    
def decode(x):
    if PY3:
        return x.decode()
    return x

def value(a):
    """
    Returns a.value if a is a Variable, or just a otherwise.
    """
    if isinstance(a, Variable):
        return a.value
    else:
        return a


# =====================================================================
# = Please don't use numpy.vectorize with these! It will leak memory. =
# =====================================================================
def logit(theta):
    return flib.logit(ravel(theta)).reshape(shape(theta))


def invlogit(ltheta):
    return flib.invlogit(ravel(ltheta)).reshape(shape(ltheta))


def stukel_invlogit(ltheta, a1, a2):
    return flib.stukel_invlogit(ravel(ltheta), a1, a2).reshape(shape(ltheta))


def stukel_logit(theta, a1, a2):
    return flib.stukel_logit(ravel(theta), a1, a2).reshape(shape(theta))


def check_list(thing, label):
    if thing is not None:
        if thing.__class__ is not list:
            return [thing]
        return thing

# TODO: Look into using numpy.core.numerictypes to do this part.


# TODO : Wrap the nd histogramming fortran function.

def safe_len(val):
    if np.isscalar(val):
        return 1
    else:
        return np.prod(np.shape(val))


def round_array(array_in):
    """
    arr_out = round_array(array_in)

    Rounds an array and recasts it to int. Also works on scalars.
    """
    if isinstance(array_in, ndarray):
        return np.round(array_in).astype(int)
    else:
        return int(np.round(array_in))

try:
    from .flib import dchdc_wrap

    def msqrt(C):
        """
        U=incomplete_chol(C)

        Computes a Cholesky factorization of C. Works for matrices that are
        positive-semidefinite as well as positive-definite, though in these
        cases the Cholesky factorization isn't unique.

        U will be upper triangular.

        This is the dchdc version. It's faster for full-rank matrices,
        but it has to compute the entire matrix.

        """
        chol = C.copy()
        piv, N = dchdc_wrap(a=chol)
        if N < 0:
            raise ValueError(
                "Matrix does not appear to be positive semidefinite")
        return asmatrix(chol[:N, argsort(piv)])

except:
    def msqrt(cov):
        """
        sig = msqrt(cov)

        Return a matrix square root of a covariance matrix. Tries Cholesky
        factorization first, and factorizes by diagonalization if that fails.
        """
        # Try Cholesky factorization
        try:
            sig = asmatrix(cholesky(cov))

        # If there's a small eigenvalue, diagonalize
        except LinAlgError:
            val, vec = eigh(cov)
            sig = np.zeros(vec.shape)
            for i in range(len(val)):
                if val[i] < 0.:
                    val[i] = 0.
                sig[:, i] = vec[:, i] * sqrt(val[i])
        return np.asmatrix(sig).T


def histogram(a, bins=10, range=None, normed=False,
              weights=None, axis=None, strategy=None):
    """histogram(a, bins=10, range=None, normed=False, weights=None, axis=None)
                                                                   -> H, dict

    Return the distribution of sample.

    :Stochastics:
      `a` : Array sample.
      `bins` : Number of bins, or an array of bin edges, in which case the
              range is not used. If 'Scott' or 'Freeman' is passed, then
              the named method is used to find the optimal number of bins.
      `range` : Lower and upper bin edges, default: [min, max].
      `normed` :Boolean, if False, return the number of samples in each bin,
              if True, return the density.
      `weights` : Sample weights. The weights are normed only if normed is
              True. Should weights.sum() not equal len(a), the total bin count
              will not be equal to the number of samples.
      `axis` : Specifies the dimension along which the histogram is computed.
              Defaults to None, which aggregates the entire sample array.
      `strategy` : Histogramming method (binsize, searchsorted or digitize).

    :Return:
      `H` : The number of samples in each bin.
        If normed is True, H is a frequency distribution.
      dict{ 'edges':      The bin edges, including the rightmost edge.
        'upper':      Upper outliers.
        'lower':      Lower outliers.
        'bincenters': Center of bins.
        'strategy': the histogramming method employed.}

    :Examples:
      >>> x = random.rand(100,10)
      >>> H, D = histogram(x, bins=10, range=[0,1], normed=True)
      >>> H2, D = histogram(x, bins=10, range=[0,1], normed=True, axis=0)

    :SeeAlso: histogramnd
    """
    weighted = weights is not None

    a = asarray(a)
    if axis is None:
        a = atleast_1d(a.ravel())
        if weighted:
            weights = atleast_1d(weights.ravel())
        axis = 0

    # Define the range
    if range is None:
        mn, mx = a.min(), a.max()
        if mn == mx:
            mn = mn - .5
            mx = mx + .5
        range = [mn, mx]

    # Find the optimal number of bins.
    if bins is None or isinstance(bins, str):
        bins = _optimize_binning(a, range, bins)

    # Compute the bin edges if they are not given explicitely.
    # For the rightmost bin, we want values equal to the right
    # edge to be counted in the last bin, and not as an outlier.
    # Hence, we shift the last bin by a tiny amount.
    if not iterable(bins):
        dr = diff(range) / bins * 1e-10
        edges = linspace(range[0], range[1] + dr, bins + 1, endpoint=True)
    else:
        edges = asarray(bins, float)

    dedges = diff(edges)
    bincenters = edges[:-1] + dedges / 2.

    # Number of bins
    nbin = len(edges) - 1

    # Measure of bin precision.
    decimal = int(-log10(dedges.min()) + 10)

    # Choose the fastest histogramming method
    even = (len(set(around(dedges, decimal))) == 1)
    if strategy is None:
        if even:
            strategy = 'binsize'
        else:
            if nbin > 30:  # approximative threshold
                strategy = 'searchsort'
            else:
                strategy = 'digitize'
    else:
        if strategy not in ['binsize', 'digitize', 'searchsort']:
            raise ValueError('Unknown histogramming strategy.', strategy)
        if strategy == 'binsize' and not even:
            raise ValueError(
                'This binsize strategy cannot be used for uneven bins.')

    # Stochastics for the fixed_binsize functions.
    start = float(edges[0])
    binwidth = float(dedges[0])

    # Looping to reduce memory usage
    block = 66600
    slices = [slice(None)] * a.ndim
    for i in arange(0, len(a), block):
        slices[axis] = slice(i, i + block)
        at = a[slices]
        if weighted:
            at = concatenate((at, weights[slices]), axis)
            if strategy == 'binsize':
                count = apply_along_axis(_splitinmiddle, axis, at,
                                         flib.weighted_fixed_binsize, start, binwidth, nbin)
            elif strategy == 'searchsort':
                count = apply_along_axis(_splitinmiddle, axis, at,
                                         _histogram_searchsort_weighted, edges)
            elif strategy == 'digitize':
                count = apply_along_axis(_splitinmiddle, axis, at,
                                         _histogram_digitize, edges, normed)
        else:
            if strategy == 'binsize':
                count = apply_along_axis(
                    flib.fixed_binsize,
                    axis,
                    at,
                    start,
                    binwidth,
                    nbin)
            elif strategy == 'searchsort':
                count = apply_along_axis(
                    _histogram_searchsort,
                    axis,
                    at,
                    edges)
            elif strategy == 'digitize':
                count = apply_along_axis(
                    _histogram_digitize, axis, at, None, edges,
                    normed)

        if i == 0:
            total = count
        else:
            total += count

    # Outlier count
    upper = total.take(array([-1]), axis)
    lower = total.take(array([0]), axis)

    # Non-outlier count
    core = a.ndim * [slice(None)]
    core[axis] = slice(1, -1)
    hist = total[core]

    if normed:
        normalize = lambda x: atleast_1d(x / (x * dedges).sum())
        hist = apply_along_axis(normalize, axis, hist)

    return hist, {'edges': edges, 'lower': lower, 'upper': upper,
                  'bincenters': bincenters, 'strategy': strategy}


def _histogram_fixed_binsize(a, start, width, n):
    """histogram_even(a, start, width, n) -> histogram

    Return an histogram where the first bin counts the number of lower
    outliers and the last bin the number of upper outliers. Works only with
    fixed width bins.

    :Stochastics:
      a : array
        Array of samples.
      start : float
        Left-most bin edge.
      width : float
        Width of the bins. All bins are considered to have the same width.
      n : int
        Number of bins.

    :Return:
      H : array
        Array containing the number of elements in each bin. H[0] is the number
        of samples smaller than start and H[-1] the number of samples
        greater than start + n*width.
    """

    return flib.fixed_binsize(a, start, width, n)


def _histogram_binsize_weighted(a, w, start, width, n):
    """histogram_even_weighted(a, start, width, n) -> histogram

    Return an histogram where the first bin counts the number of lower
    outliers and the last bin the number of upper outliers. Works only with
    fixed width bins.

    :Stochastics:
      a : array
        Array of samples.
      w : array
        Weights of samples.
      start : float
        Left-most bin edge.
      width : float
        Width of the bins. All bins are considered to have the same width.
      n : int
        Number of bins.

    :Return:
      H : array
        Array containing the number of elements in each bin. H[0] is the number
        of samples smaller than start and H[-1] the number of samples
        greater than start + n*width.
    """
    return flib.weighted_fixed_binsize(a, w, start, width, n)


def _histogram_searchsort(a, bins):
    n = sort(a).searchsorted(bins)
    n = concatenate([n, [len(a)]])
    count = concatenate([[n[0]], n[1:] - n[:-1]])
    return count


def _histogram_searchsort_weighted(a, w, bins):
    i = sort(a).searchsorted(bins)
    sw = w[argsort(a)]
    i = concatenate([i, [len(a)]])
    n = concatenate([[0], sw.cumsum()])[i]
    count = concatenate([[n[0]], n[1:] - n[:-1]])
    return count


def _splitinmiddle(x, function, *args, **kwds):
    x1, x2 = hsplit(x, 2)
    return function(x1, x2, *args, **kwds)


def _histogram_digitize(a, w, edges, normed):
    """Internal routine to compute the 1d weighted histogram for uneven bins.
    a: sample
    w: weights
    edges: bin edges
    weighted: Means that the weights are appended to array a.
    Return the bin count or frequency if normed.
    """
    weighted = w is not None
    nbin = edges.shape[0] + 1
    if weighted:
        count = zeros(nbin, dtype=w.dtype)
        if normed:
            count = zeros(nbin, dtype=float)
            w = w / w.mean()
    else:
        count = zeros(nbin, int)

    binindex = digitize(a, edges)

    # Count the number of identical indices.
    flatcount = bincount(binindex, w)

    # Place the count in the histogram array.
    count[:len(flatcount)] = flatcount

    return count


def _optimize_binning(x, range, method='Freedman'):
    """Find the optimal number of bins.
    Available methods : Freedman, Scott
    """
    N = x.shape[0]
    if method.lower() == 'freedman':
        s = sort(x)
        IQR = s[int(N * .75)] - s[int(
            N * .25)]  # Interquantile range (75% -25%)
        width = 2 * IQR * N ** (-1. / 3)

    elif method.lower() == 'scott':
        width = 3.49 * x.std() * N ** (-1. / 3)
    else:
        raise ValueError('Method must be Scott or Freedman', method)
    return int(diff(range) / width)


def normcdf(x, log=False):
    """Normal cumulative density function."""
    y = np.atleast_1d(x).copy()
    flib.normcdf(y)
    if log:
        if (y>0).all():
            return np.log(y)
        return -np.inf
    return y


def lognormcdf(x, mu, tau):
    """Log-normal cumulative density function"""
    x = np.atleast_1d(x)
    return np.array(
        [0.5 * (1 - flib.derf(-(np.sqrt(tau / 2)) * (np.log(y) - mu))) for y in x])


def invcdf(x):
    """Inverse of normal cumulative density function."""
    x_flat = np.ravel(x)
    x_trans = np.array([flib.ppnd16(y, 1) for y in x_flat])
    return np.reshape(x_trans, np.shape(x))


def ar1_gen(rho, mu, sigma, size=1):
    """Create an autoregressive series of order one AR(1) generator.

    .. math::
        X_t = \mu_t + \rho (X_{t-1}-\mu_{t-1} + \epsilon_t

    If mu is a sequence and size > len(mu), the algorithm loops through
    mu.

    :Stochastics:
        rho : scalar in [0,1]
        mu : scalar or sequence
        sigma : scalar > 0
        size : integer
    """
    mu = np.asarray(mu, float)
    mu = np.resize(mu, size)
    r = mu.copy()
    r += np.random.randn(size) * sigma
    r[0] = np.random.randn(1) * sigma / np.sqrt(1 - rho ** 2)
    i = 0
    while True:
        yield r[i]
        i += 1
        if i == size:
            break
        r[i] += rho * (r[i - 1] - mu[i - 1])


def ar1(rho, mu, sigma, size=1):
    """Return an autoregressive series of order one AR(1).

    .. math::
        X_t = \mu_t + \rho (X_{t-1}-\mu_{t-1} + \epsilon_t

    If mu is a sequence and size > len(mu), the algorithm loops through
    mu.

    :Stochastics:
        rho : scalar in [0,1]
        mu : scalar or sequence
        sigma : scalar > 0
        size : integer
    """
    return np.array([x for x in ar1_gen(rho, mu, sigma, size)])


def autocorr(x, lag=1):
    """Sample autocorrelation at specified lag.
    The autocorrelation is the correlation of x_i with x_{i+lag}.
    """

    if not lag:
        return 1
    if lag < 0:
        return
    # x = np.squeeze(asarray(x))
    #     mu = x.mean()
    #     v = x.var()
    #     return ((x[:-lag]-mu)*(x[lag:]-mu)).sum()/v/(len(x) - lag)
    S = autocov(x, lag)
    return S[0, 1] / sqrt(prod(diag(S)))


def autocov(x, lag=1):
    """
    Sample autocovariance at specified lag.
    The autocovariance is a 2x2 matrix with the variances of
    x[:-lag] and x[lag:] in the diagonal and the autocovariance
    on the off-diagonal.
    """

    if not lag:
        return 1
    if lag < 0:
        return
    return cov(x[:-lag], x[lag:], bias=1)


def trace_generator(trace, start=0, stop=None, step=1):
    """Return a generator returning values from the object's trace.

    Ex:
    T = trace_generator(theta.trace)
    T.next()
    for t in T:...
    """
    i = start
    stop = stop or np.inf
    size = min(trace.length(), stop)
    while i < size:
        index = slice(i, i + 1)
        yield trace.gettrace(slicing=index)[0]
        i += step


def draw_random(obj, **kwds):
    """Draw random variates from obj.random method.

    If the object has parents whose value must be updated, use
    parent_name=trace_generator_function.

    Ex:
    R = draw_random(theta, beta=pymc.utils.trace_generator(beta.trace))
    R.next()
    """
    while True:
        for k, v in six.iteritems(kwds):
            obj.parents[k] = v.next()
        yield obj.random()


def rec_getattr(obj, attr):
    """Get object's attribute. May use dot notation.

    >>> class C(object): pass
    >>> a = C()
    >>> a.b = C()
    >>> a.b.c = 4
    >>> rec_getattr(a, 'b.c')
    4
    """
    return reduce(getattr, attr.split('.'), obj)


def rec_setattr(obj, attr, value):
    """Set object's attribute. May use dot notation.

    >>> class C(object): pass
    >>> a = C()
    >>> a.b = C()
    >>> a.b.c = 4
    >>> rec_setattr(a, 'b.c', 2)
    >>> a.b.c
    2
    """
    attrs = attr.split('.')
    setattr(reduce(getattr, attrs[:-1], obj), attrs[-1], value)


def hpd(x, alpha):
    """Calculate highest posterior density (HPD) of array for given alpha. The HPD is the
    minimum width Bayesian credible interval (BCI).

    :Arguments:
      x : Numpy array
          An array containing MCMC samples
      alpha : float
          Desired probability of type I error

    """

    # Make a copy of trace
    x = x.copy()

    # For multivariate node
    if x.ndim > 1:

        # Transpose first, then sort
        tx = tr(x, list(range(x.ndim)[1:]) + [0])
        dims = shape(tx)

        # Container list for intervals
        intervals = np.resize(0.0, (2,) + dims[:-1])

        for index in make_indices(dims[:-1]):

            try:
                index = tuple(index)
            except TypeError:
                pass

            # Sort trace
            sx = sort(tx[index])

            # Append to list
            intervals[0][index], intervals[1][index] = calc_min_interval(sx, alpha)
        
        # Transpose back before returning
        return array(intervals)

    else:
        # Sort univariate node
        sx = sort(x)

        return array(calc_min_interval(sx, alpha))


def make_indices(dimensions):
    # Generates complete set of indices for given dimensions

    level = len(dimensions)

    if level == 1:
        return range(dimensions[0])

    indices = [[]]

    while level:

        _indices = []

        for j in range(dimensions[level - 1]):

            _indices += [[j] + i for i in indices]

        indices = _indices

        level -= 1

    try:
        return [tuple(i) for i in indices]
    except TypeError:
        return indices


def calc_min_interval(x, alpha):
    """Internal method to determine the minimum interval of
    a given width

    Assumes that x is sorted numpy array.
    """

    n = len(x)
    cred_mass = 1.0 - alpha

    interval_idx_inc = int(np.floor(cred_mass * n))
    n_intervals = n - interval_idx_inc
    interval_width = x[interval_idx_inc:] - x[:n_intervals]

    if len(interval_width) == 0:
        print_('Too few elements for interval calculation')
        return [None, None]

    min_idx = np.argmin(interval_width)
    hdi_min = x[min_idx]
    hdi_max = x[min_idx + interval_idx_inc]
    return [hdi_min, hdi_max]


def quantiles(x, qlist=(2.5, 25, 50, 75, 97.5)):
    """Returns a dictionary of requested quantiles from array

    :Arguments:
      x : Numpy array
          An array containing MCMC samples
      qlist : tuple or list
          A list of desired quantiles (defaults to (2.5, 25, 50, 75, 97.5))

    """

    # Make a copy of trace
    x = x.copy()

    # For multivariate node
    if x.ndim > 1:
        # Transpose first, then sort, then transpose back
        sx = sort(x.T).T
    else:
        # Sort univariate node
        sx = sort(x)

    try:
        # Generate specified quantiles
        quants = [sx[int(len(sx) * q / 100.0)] for q in qlist]

        return dict(zip(qlist, quants))

    except IndexError:
        print_("Too few elements for quantile calculation")


def coda_output(pymc_object, name=None, chain=-1):
    """Generate output files that are compatible with CODA

    :Arguments:
      pymc_object : Model or Node
          A PyMC object containing MCMC output.

    """

    print_()
    print_("Generating CODA output")
    print_('=' * 50)

    if name is None:
        name = pymc_object.__name__

    # Open trace file
    trace_file = open(name + '_coda.out', 'w')

    # Open index file
    index_file = open(name + '_coda.ind', 'w')

    variables = [pymc_object]
    if hasattr(pymc_object, 'stochastics'):
        variables = pymc_object.stochastics

    # Initialize index
    index = 1

    # Loop over all parameters
    for v in variables:

        vname = v.__name__
        print_("Processing", vname)
        try:
            index = _process_trace(
                trace_file,
                index_file,
                v.trace(chain=chain),
                vname,
                index)
        except TypeError:
            pass

    # Close files
    trace_file.close()
    index_file.close()

# Lazy shortcut
coda = coda_output


def _process_trace(trace_file, index_file, trace, name, index):
    """Support function for coda_output(); writes output to files"""

    if ndim(trace) > 1:
        trace = swapaxes(trace, 0, 1)
        for i, seq in enumerate(trace):
            _name = '%s_%s' % (name, i)
            index = _process_trace(trace_file, index_file, seq, _name, index)
    else:
        index_buffer = '%s\t%s\t' % (name, index)
        for i, val in enumerate(trace):
            trace_file.write('%s\t%s\r\n' % (i + 1, val))
            index += 1
        index_file.write('%s%s\r\n' % (index_buffer, index - 1))

    return index


def log_difference(lx, ly):
    """Returns log(exp(lx) - exp(ly)) without leaving log space."""

    # Negative log of double-precision infinity
    li = -709.78271289338397
    diff = ly - lx
    # Make sure log-difference can succeed
    if np.any(diff >= 0):
        raise ValueError(
            'Cannot compute log(x-y), because y>=x for some elements.')
    # Otherwise evaluate log-difference
    return lx + np.log(1. - np.exp(diff))


def getInput():
    """Read the input buffer without blocking the system."""
    input = ''

    if sys.platform == 'win32':
        import msvcrt
        if msvcrt.kbhit():  # Check for a keyboard hit.
            input += msvcrt.getch()
            print_(input)
        else:
            time.sleep(.1)

    else:  # Other platforms
        # Posix will work with sys.stdin or sys.stdin.fileno()
        # Mac needs the file descriptor.
        # This solution does not work for windows since select
        # expects a socket, and I have no idea how to create a
        # socket from standard input.
        sock = sys.stdin.fileno()

        # select(rlist, wlist, xlist, timeout)
        while len(select.select([sock], [], [], 0.1)[0]) > 0:
            input += decode(os.read(sock, 4096))

    return input


def crawl_dataless(sofar, gens):
    """
    Crawls out from v to find the biggest dataless submodel containing v.
    TODO: Let MCMC start the crawl from its last generation. It doesn't
    matter that there won't be one contiguous group.
    """
    new_gen = set([])
    all_ext_parents = set()
    for s in gens[-1]:
        all_ext_parents.update(s.extended_parents)

    for p in all_ext_parents:
        if p._random is not None and not p.observed:
            if len(p.extended_children - sofar) == 0:
                new_gen.add(p)
    if len(new_gen) == 0:
        return sofar, gens
    else:
        sofar |= new_gen
        gens.append(new_gen)
        return crawl_dataless(sofar, gens)


def find_generations(container, with_data=False):
    """
    A generation is the set of stochastic variables that only has parents in
    previous generations.
    """

    generations = []

    # Find root generation
    generations.append(set())
    all_children = set()
    if with_data:
        stochastics_to_iterate = container.stochastics | container.observed_stochastics
    else:
        stochastics_to_iterate = container.stochastics
    for s in stochastics_to_iterate:
        all_children.update(s.extended_children & stochastics_to_iterate)
    generations[0] = stochastics_to_iterate - all_children

    # Find subsequent _generations
    children_remaining = True
    gen_num = 0
    while children_remaining:
        gen_num += 1

        # Find children of last generation
        generations.append(set())
        for s in generations[gen_num - 1]:
            generations[gen_num].update(
                s.extended_children & stochastics_to_iterate)

        # Take away stochastics that have parents in the current generation.
        thisgen_children = set()
        for s in generations[gen_num]:
            thisgen_children.update(
                s.extended_children & stochastics_to_iterate)
        generations[gen_num] -= thisgen_children

        # Stop when no subsequent _generations remain
        if len(thisgen_children) == 0:
            children_remaining = False
    return generations


def append(nodelist, node, label=None, sep='_'):
    """
    Append function to automate the naming of list elements in Containers.

    :Arguments:
        - `nodelist` : List containing nodes for Container.
        - `node` : Node to be added to list.
        - `label` : Label to be appended to list (If not passed,
        defaults to element number).
        - `sep` : Separator character for label (defaults to underscore).

    :Return:
        - `nodelist` : Passed list with node added.

    """

    nname = node.__name__

    # Determine label
    label = label or len(nodelist)

    # Look for separator at the end of name
    ind = nname.rfind(sep)

    # If there is no separator, we will remove last character and
    # replace with label.
    node.__name__ = nname[:ind] + sep + str(label)

    nodelist.append(node)

    return nodelist


# deterministic related utilities
def find_element(names, modules, error_on_fail):
    element = None
    found = False

    if isinstance(names, str):
        names = [names]

    if isinstance(modules, dict) or isinstance(modules, types.ModuleType):
        modules = [modules]

    for module in modules:

        if isinstance(module, types.ModuleType):
            module = copy(module.__dict__)
        elif isinstance(module, dict):
            module = copy(module)
        else:
            raise AttributeError

        for name in names:
            try:
                function = module[name]
                found = True
            except KeyError:
                pass

    if not found and error_on_fail:
        raise NameError(
            "no function or variable " + str(
                names) + " in " + str(
                modules))

    return function



# Copyright (c) Anand Patil, 2007

# Where matrix evaluations can be done in chunks, the chunk size will be
# kept below this limit.
chunksize = 1e8

__modules__ = ['GPutils',
               'Mean',
               'Covariance',
               'BasisCovariance',
               'FullRankCovariance',
               'NearlyFullRankCovariance',
               'Realization',
               'cov_funs',
               'gp_submodel',
               'step_methods']

__optmodules__ = ['gpplots']

from .GPutils import *
from .Mean import *
from .Covariance import *
from .BasisCovariance import *
from .FullRankCovariance import *
from .NearlyFullRankCovariance import *
from .Realization import *
from .cov_funs import *
from .gp_submodel import *
from .step_methods import *

try:
    from . import gpplots
except ImportError:
    pass

try:
    import SparseCovariance
except ImportError:
    pass



from pymc import *
from numpy import ones, array

# Samples for each dose level
n = 5 * ones(4, dtype=int)
# Log-dose
dose = array([-.86, -.3, -.05, .73])

# Logit-linear model parameters
alpha = Normal('alpha', 0, 0.01)
beta = Normal('beta', 0, 0.01)

# Calculate probabilities of death
theta = Lambda('theta', lambda a=alpha, b=beta, d=dose: invlogit(a + b * d))

# Data likelihood
deaths = Binomial(
    'deaths',
    n=n,
    p=theta,
    value=array([0,
                 1,
                 3,
                 5],
                dtype=float),
    observed=True)

# Calculate LD50
LD50 = Lambda('LD50', lambda a=alpha, b=beta: -a / b)



from pymc.gp.cov_funs import *
from numpy import *
from copy import copy
from pymc.gp import *

N = 25

# Generate mean


def quadfun(x, a, b, c):
    return (a * x ** 2 + b * x + c)

M = Mean(eval_fun=quadfun, a=1., b=.5, c=2.)

# Generate basis covariance
coef_cov = ones(2 * N + 1, dtype=float)
for i in xrange(1, len(coef_cov)):
    coef_cov[i] = 1. / int(((i + 1) / 2)) ** 2.5

basis = fourier_basis([N])

C = SeparableBasisCovariance(basis, coef_cov, xmin=[-2.], xmax=[2.])

obs_x = array([-.5, .5])
V = array([.002, .002])
data = array([3.1, 2.9])

observe(M=M,
        C=C,
        obs_mesh=obs_x,
        obs_V=V,
        obs_vals=data)

if __name__ == '__main__':
    from pylab import *

    close('all')
    x = arange(-1., 1., .01)

    figure()
    # Plot the covariance function
    subplot(1, 2, 1)

    contourf(
        x,
        x,
        C(x,
            x).view(ndarray),
        origin='lower',
        extent=(-1.,
                1.,
                -1.,
                1.),
        cmap=cm.bone)

    xlabel('x')
    ylabel('y')
    title('C(x,y)')
    axis('tight')
    colorbar()

    # Plot a slice of the covariance function
    subplot(1, 2, 2)

    plot(x, C(x, 0.).view(ndarray).ravel(), 'k-')

    xlabel('x')
    ylabel('C(x,.5)')
    title('A slice of C')

    figure()
    plot_envelope(M, C, x)
    for i in range(3):
        f = Realization(M, C)
        plot(x, f(x))
        title('Three realizations')

    plot(obs_x, data, 'k.', markersize=16)

    # show()



"""
Test interactive sampler
"""

# TODO: Make real test case.

from pymc import MCMC
from pymc.examples import disaster_model
import os
import nose


def test_interactive():
    S = MCMC(disaster_model)
    S.isample(
        200,
        100,
        2,
        out=open(
            'testresults/interactive.log',
            'w'),
        progress_bar=0)


if __name__ == '__main__':
    nose.runmodule()



"""The point of Container.py is to provide a function Container which converts
any old thing A to thing B which looks and acts just like A, but it has a
'value' attribute. B.value looks and acts just like A but every variable
'inside' B has been replaced by its value. Examples:

    class MyObject(object):
        def __init__(self):
            self.x = Uninformative('x',0)
            self.y = 3

    A = MyObject()
    B = Container(A)
    B.x
    B.value.x


    A = [Uninformative('x',0), 3]
    B = Container(A)
    B
    B.value

Should work even with nested inputs:

    class MyObject(object):
        def __init__(self):
            self.x = [Uninformative('x',0), 5]
            self.y = 3

    A = MyObject()
    B = Container(A)


In addition, container objects file away the objects they contain into the
following sets: stochastics, deterministics, variables, nodes, containers, data, step methods.
These flattened representations are useful for things like cache checking.
"""

from .Node import Node, ContainerBase, Variable, StochasticBase, DeterministicBase, PotentialBase, ContainerRegistry
from copy import copy
from numpy import ndarray, array, zeros, shape, arange, where, dtype, Inf
from .Container_values import LCValue, DCValue, ACValue, OCValue
from types import ModuleType
import pdb

from . import six
xrange = six.moves.xrange

__all__ = [
    'Container',
    'DictContainer',
    'TupleContainer',
    'ListContainer',
    'SetContainer',
    'ObjectContainer',
    'ArrayContainer']


def filter_dict(obj):
    filtered_dict = {}
    for item in six.iteritems(obj.__dict__):
        if isinstance(item[1], Node) or isinstance(item[1], ContainerBase):
            filtered_dict[item[0]] = item[1]
    return filtered_dict


def Container(*args):
    """
    C = Container(iterable)
    C = Container(module)
    C = Container(object)
    C = Container(obj_1, obj_2, obj_3, ...)

    Wraps an iterable object (currently a list, set, tuple, dictionary
    or ndarray), or a module or other object, or just a sequence of objects,
    in a subclass of ContainerBase and returns it.

    Iterable subclasses of ContainerBase strive to emulate the iterables they
    wrap, with one important difference: They have a value attribute.
    A container's value attribute behaves like the container itself, but
    it replaces every PyMC variable it contains with that variable's value.

    Example:

        @stochastic
        def A(value=0., mu=3, tau=2):
            return normal_like(value, mu, tau)

        C = Container([A, 15.2])

        will yield the following:
        C[0] = A
        C.value[0] = A.value
        C[1] = C.value[1] = 15.2


    The primary reason containers exist is to allow nodes to have large
    sets of parents without the need to refer to each of the parents by name.
    Example:

        x = []

        @stochastic
        def x_0(value=0, mu=0, tau=2):
            return normal_like(value, mu, tau)

        x.append(x_0)
        last_x = x_0

        for i in range(1,N):
            @stochastic
            def x_now(value=0, mu = last_x, tau=2):
                return normal_like(value, mu, tau)

            x_now.__name__ = 'x[%i]' % i
            last_x = x_now

            x.append(x_now)

        @stochastic
        def y(value=0, mu = x, tau = 100):

            mean_sum = 0
            for i in range(len(mu)):
                mean_sum = mean_sum + mu[i]

            return normal_like(value, mean_sum, tau)

    x.value will be passed into y's log-probability function as argument mu,
    so mu[i] will return x.value[i] = x[i].value. Stochastic y
    will cache the values of each element of x, and will evaluate whether it
    needs to recompute based on all of them.

    :SeeAlso:
      ListContainer, TupleContainer, SetContainer, ArrayContainer, DictContainer,
      ObjectContainer
    """

    if len(args) == 1:
        iterable = args[0]
    else:
        iterable = args

    if isinstance(iterable, ContainerBase):
        return iterable

    for container_class, containing_classes in ContainerRegistry:
        if any([isinstance(iterable, containing_class)
                for containing_class in containing_classes]):
            return container_class(iterable)

    # Wrap mutable objects
    # if hasattr(iterable, '__dict__'):
    #     return ObjectContainer(iterable.__dict__)

    # Otherwise raise an error.
    raise ValueError(
        'No container classes available for class ' +
        iterable.__class__.__name__ +
        ', see Container.py for examples on how to write one.')


class _A(object):
    pass
dict_proxy_type = type(_A.__dict__)
del _A


def file_items(container, iterable):
    """
    Files away objects into the appropriate attributes of the container.
    """

    # container._value = copy(iterable)

    container.nodes = set()
    container.variables = set()
    container.deterministics = set()
    container.stochastics = set()
    container.potentials = set()
    container.observed_stochastics = set()

    # containers needs to be a list to hold unhashable items.
    container.containers = []

    i = -1

    for item in iterable:

        # If this is a dictionary, switch from key to item.
        if isinstance(iterable, (dict, dict_proxy_type)):
            key = item
            item = iterable[key]
        # Item counter
        else:
            i += 1

        # If the item isn't iterable, file it away.
        if isinstance(item, Variable):
            container.variables.add(item)
            if isinstance(item, StochasticBase):
                if item.observed or not getattr(item, 'mask', None) is None:
                    container.observed_stochastics.add(item)
                if not item.observed:
                    container.stochastics.add(item)
            elif isinstance(item, DeterministicBase):
                container.deterministics.add(item)
        elif isinstance(item, PotentialBase):
            container.potentials.add(item)

        elif isinstance(item, ContainerBase):
            container.assimilate(item)
            container.containers.append(item)

        # Wrap internal containers
        elif hasattr(item, '__iter__'):

            # If this is a non-object-valued ndarray, don't container-ize it.
            if isinstance(item, ndarray):
                if item.dtype != dtype('object'):
                    continue

            # If the item is iterable, wrap it in a container. Replace the item
            # with the wrapped version.
            try:
                new_container = Container(item)
            except:
                continue

            # Update all of container's variables, potentials, etc. with the new wrapped
            # iterable's. This process recursively unpacks nested iterables.
            container.assimilate(new_container)

            if isinstance(container, dict):
                container.replace(key, new_container)
            elif isinstance(container, tuple):
                return container[:i] + (new_container,) + container[i + 1:]
            else:
                container.replace(item, new_container, i)

    container.nodes = container.potentials | container.variables

    # 'Freeze' markov blanket, moral neighbors, coparents of all constituent stochastics
    # for future use
    for attr in ['moral_neighbors', 'markov_blanket', 'coparents']:
        setattr(container, attr, {})
    for s in container.stochastics:
        for attr in ['moral_neighbors', 'markov_blanket', 'coparents']:
            getattr(container, attr)[s] = getattr(s, attr)

value_doc = 'A copy of self, with all variables replaced by their values.'


def sort_list(container, _value):
    val_ind = []
    val_obj = []
    nonval_ind = []
    nonval_obj = []
    for i in xrange(len(_value)):
        obj = _value[i]
        if isinstance(obj, Variable) or isinstance(obj, ContainerBase):
            val_ind.append(i)
            val_obj.append(obj)
        else:
            nonval_ind.append(i)
            nonval_obj.append(obj)
    # In case val_obj is only a single array, avert confusion.
    # Leave this even though it's confusing!
    val_obj.append(None)
    nonval_obj.append(None)
    container.n_val = len(val_ind)
    container.n_nonval = len(nonval_ind)
    container.val_ind = array(val_ind, dtype='int32')
    container.val_obj = val_obj
    container.nonval_ind = array(nonval_ind, dtype='int32')
    container.nonval_obj = array(nonval_obj, dtype=object)
    container.LCValue = LCValue(container)


class SetContainer(ContainerBase, frozenset):

    """
    SetContainers are containers that wrap sets.

    :Parameters:
      iterable : set.

    :Attributes:
      value : set
        A copy of self, with all variables replaced with their values.
      nodes : set
        All the stochastics, deterministics and potentials self contains.
      deterministics : set
        All the deterministics self contains.
      stochastics : set
        All the stochastics self contains with observed=False.
      potentials : set
        All the potentials self contains.
      observed_stochastics : set
        All the stochastics self contains with observed=True.
      containers : list
        All the containers self contains.

    :Note:
      - nodes, deterministics, etc. include all the objects in nested
        containers.
      - value replaces objects in nested containers.

    :SeeAlso:
      Container, ListContainer, DictContainer, ArrayContainer, TupleContainer,
      ObjectContainer
    """
    register = True
    change_methods = []
    containing_classes = [set, frozenset]

    def __init__(self, iterable):
        self.new_iterable = set(iterable)
        file_items(self, self.new_iterable)
        ContainerBase.__init__(self, self.new_iterable)
        self._value = list(self)
        sort_list(self, self._value)

    def replace(self, item, new_container, i):
        self.new_iterable.discard(item)
        self.new_iterable.add(new_container)

    def get_value(self):
        self.LCValue.run()
        return set(self._value)

    value = property(fget=get_value, doc=value_doc)


class TupleContainer(ContainerBase, tuple):

    """
    TupleContainers are containers that wrap tuples.

    :Parameters:
      iterable : tuple.

    :Attributes:
      value : tuple
        A copy of self, with all variables replaced with their values.
      nodes : set
        All the stochastics, deterministics and potentials self contains.
      deterministics : set
        All the deterministics self contains.
      stochastics : set
        All the stochastics self contains with observed=False.
      potentials : set
        All the potentials self contains.
      observed_stochastics : set
        All the stochastics self contains with observed=True.
      containers : list
        All the containers self contains.

    :Note:
      - nodes, deterministics, etc. include all the objects in nested
        containers.
      - value replaces objects in nested containers.

    :SeeAlso:
      Container, ListContainer, DictContainer, ArrayContainer, SetContainer,
      ObjectContainer
    """
    register = True
    change_methods = []
    containing_classes = [tuple]

    def __init__(self, iterable):
        new_tup = file_items(self, iterable)
        if len(self.containers) > 0:
            raise NotImplementedError("""We have not figured out how to satisfactorily implement nested TupleContainers.
The reason is there is no way to change an element of a tuple after it has been created.
Even the Python-C API makes this impossible by checking that a tuple is new
before allowing you to change one of its elements.""")
        ContainerBase.__init__(self, iterable)
        file_items(self, iterable)
        self._value = list(self)
        sort_list(self, self._value)

    def replace(self, item, new_container, i):
        list.__setitem__(self, i, new_container)

    def get_value(self):
        self.LCValue.run()
        return tuple(self._value)

    value = property(fget=get_value, doc=value_doc)


class ListContainer(ContainerBase, list):

    """
    ListContainers are containers that wrap lists.

    :Parameters:
      iterable : list.

    :Attributes:
      value : list
        A copy of self, with all variables replaced with their values.
      nodes : set
        All the stochastics, deterministics and potentials self contains.
      deterministics : set
        All the deterministics self contains.
      stochastics : set
        All the stochastics self contains with observed=False.
      potentials : set
        All the potentials self contains.
      observed_stochastics : set
        All the stochastics self contains with observed=True.
      containers : list
        All the containers self contains.

    :Note:
      - nodes, deterministics, etc. include all the objects in nested
        containers.
      - value replaces objects in nested containers.

    :SeeAlso:
      Container, TupleContainer, DictContainer, ArrayContainer, SetContainer,
      ObjectContainer
    """
    change_methods = [
        '__setitem__',
        '__delitem__',
        '__setslice__',
        '__delslice__',
        '__iadd__',
        '__imul__',
        'append',
        'extend',
        'insert',
        'pop',
        'remove',
        'reverse',
        'sort']
    containing_classes = [list]
    register = True

    def __init__(self, iterable):
        list.__init__(self, iterable)
        ContainerBase.__init__(self, iterable)
        file_items(self, iterable)
        self._value = list(self)
        sort_list(self, self._value)

    def replace(self, item, new_container, i):
        list.__setitem__(self, i, new_container)

    def get_value(self):
        self.LCValue.run()
        return self._value

    value = property(fget=get_value, doc=value_doc)


class DictContainer(ContainerBase, dict):

    """
    DictContainers are containers that wrap dictionaries.
    Modules are converted into DictContainers, and variables' and potentials'
    Parents objects are DictContainers also.

    :Parameters:
      iterable : dictionary or object with a __dict__.

    :Attributes:
      value : dictionary
        A copy of self, with all variables replaced with their values.
      nodes : set
        All the stochastics, deterministics and potentials self contains.
      deterministics : set
        All the deterministics self contains.
      stochastics : set
        All the stochastics self contains with observed=False.
      potentials : set
        All the potentials self contains.
      observed_stochastics : set
        All the stochastics self contains with observed=True.
      containers : list
        All the containers self contains.

    :Note:
      - nodes, deterministics, etc. include all the objects in nested
        containers.
      - value replaces objects in nested containers.

    :SeeAlso:
      Container, ListContainer, TupleContainer, ArrayContainer, SetContainer,
      ObjectContainer
    """
    change_methods = [
        '__setitem__',
        '__delitem__',
        'clear',
        'pop',
        'popitem',
        'update']
    containing_classes = [dict]
    register = True

    def __init__(self, iterable):
        dict.__init__(self, iterable)
        ContainerBase.__init__(self, iterable)
        self._value = copy(iterable)
        file_items(self, iterable)

        self.val_keys = []
        self.val_obj = []
        self.nonval_keys = []
        self.nonval_obj = []
        self._value = {}
        for key, obj in six.iteritems(self):
            if isinstance(obj, Variable) or isinstance(obj, ContainerBase):
                self.val_keys.append(key)
                self.val_obj.append(obj)
            else:
                self.nonval_keys.append(key)
                self.nonval_obj.append(obj)
        # In case val_obj is only a single array, avert confusion.
        # Leave this even though it's confusing!
        self.val_obj.append(None)
        self.nonval_obj.append(None)

        self.n_val = len(self.val_keys)
        self.val_keys = array(self.val_keys, dtype=object)
        # self.val_obj = array(self.val_obj, dtype=object)
        self.n_nonval = len(self.nonval_keys)
        self.nonval_keys = array(self.nonval_keys, dtype=object)
        self.nonval_obj = array(self.nonval_obj, dtype=object)
        self.DCValue = DCValue(self)

    def replace(self, key, new_container):
        dict.__setitem__(self, key, new_container)

    def get_value(self):
        # DCValue(self)
        self.DCValue.run()
        return self._value

    value = property(fget=get_value, doc=value_doc)


def conservative_update(obj, dict):
    for k in dict:
        if not hasattr(obj, k):
            try:
                setattr(obj, k, dict[k])
            except:
                pass


class ObjectContainer(ContainerBase):

    """
    ObjectContainers wrap non-iterable objects.

    Contents of the input iterable, or attributes of the input object,
    are exposed as attributes of the object.

    :Parameters:
      iterable : dictionary or object with a __dict__.

    :Attributes:
      value : object
        A copy of self, with all variables replaced with their values.
      nodes : set
        All the stochastics, deterministics and potentials self contains.
      deterministics : set
        All the deterministics self contains.
      stochastics : set
        All the stochastics self contains with observed=False.
      potentials : set
        All the potentials self contains.
      observed_stochastics : set
        All the stochastics self contains with observed=True.
      containers : list
        All the containers self contains.

    :Note:
      - nodes, deterministics, etc. include all the objects in nested
        containers.
      - value replaces objects in nested containers.

    :SeeAlso:
      Container, ListContainer, DictContainer, ArrayContainer, SetContainer,
      TupleContainer
    """
    register = False

    def __init__(self, input):

        if isinstance(input, dict):
            input_to_file = input
            conservative_update(self, input_to_file)
            # self.__dict__.update(input_to_file)

        elif hasattr(input, '__iter__'):
            input_to_file = input

        else:  # Modules, objects, etc.
            input_to_file = input.__dict__
            conservative_update(self, input_to_file)
            # self.__dict__.update(input_to_file)

        dictpop = copy(self.__dict__)
        if 'self' in dictpop:
            dictpop.pop('self')

        self._dict_container = DictContainer(dictpop)
        file_items(self, input_to_file)

        self._value = copy(self)
        ContainerBase.__init__(self, input)
        self.OCValue = OCValue(self)

    def replace(self, item, new_container, key):
        dict.__setitem__(self.__dict__, key, new_container)

    def _get_value(self):
        self.OCValue.run()
        return self._value
    value = property(fget=_get_value, doc=value_doc)


class ArrayContainer(ContainerBase, ndarray):

    """
    ArrayContainers wrap Numerical Python ndarrays. These are full
    ndarray subclasses, and should support all of ndarrays'
    functionality.

    :Parameters:
      iterable : array.

    :Attributes:
      value : array.
        A copy of self, with all variables replaced with their values.
      nodes : set
        All the stochastics, deterministics and potentials self contains.
      deterministics : set
        All the deterministics self contains.
      stochastics : set
        All the stochastics self contains with observed=False.
      potentials : set
        All the potentials self contains.
      observed_stochastics : set
        All the stochastics self contains with observed=True.
      containers : list
        All the containers self contains.

    :Note:
      - nodes, deterministics, etc. include all the objects in nested
        containers.
      - value replaces objects in nested containers.

    :SeeAlso:
      Container, ListContainer, DictContainer, ObjectContainer, SetContainer,
      TupleContainer
    """

    register = True
    change_methods = []
    containing_classes = [ndarray]

    def __new__(subtype, array_in):
        if not array_in.dtype == dtype('object'):
            raise ValueError(
                'Cannot create container from array whose dtype is not object.')

        C = array(array_in, copy=True).view(subtype)
        C_ravel = C.ravel()
        ContainerBase.__init__(C, array_in)

        # Sort out contents and wrap internal containers.
        file_items(C, C_ravel)
        C._value = C.copy()
        C._ravelledvalue = C._value.ravel()

        # An array range to keep around.
        C.iterrange = arange(len(C_ravel))

        val_ind = []
        val_obj = []
        nonval_ind = []
        nonval_obj = []
        for i in xrange(len(C_ravel)):
            obj = C_ravel[i]
            if isinstance(obj, Variable) or isinstance(obj, ContainerBase):
                val_ind.append(i)
                val_obj.append(obj)
            else:
                nonval_ind.append(i)
                nonval_obj.append(obj)
        val_obj.append(None)
        C.val_ind = array(val_ind, dtype='int32')
        C.val_obj = val_obj
        C.n_val = len(val_ind)
        nonval_obj.append(None)
        C.nonval_ind = array(nonval_ind, dtype='int32')
        C.nonval_obj = array(nonval_obj, dtype=object)
        C.n_nonval = len(nonval_ind)

        C.flags['W'] = False
        C.ACValue = ACValue(C)

        return C

    def replace(self, item, new_container, i):
        ndarray.__setitem__(self.ravel(), i, new_container)

    # This method converts self to self.value.
    def get_value(self):
        self.ACValue.run()
        return self._value

    value = property(fget=get_value, doc=value_doc)



# Copyright (c) Anand Patil, 2007

__docformat__ = 'reStructuredText'

__all__ = ['GaussianProcess', 'GPEvaluation', 'GPSubmodel']

import pymc as pm
import copy
import numpy as np
from .Realization import Realization


__all__ = ['GaussianProcess', 'GPEvaluation', 'GPSubmodel']


def gp_logp(x, M, C, mesh, f_eval, M_obs, C_obs):
    raise TypeError('GP objects have no logp function')


def gp_rand(M, C, mesh, f_eval, M_obs, C_obs, size=None):
    # M and C are input pre-observed, so no need to
    out = pm.gp.Realization(M_obs, C_obs)
    out.x_sofar = mesh
    out.f_sofar = f_eval
    out.M = M
    out.C = C
    return out


class GaussianProcess(pm.Stochastic):

    """
    G=GaussianProcess(name, submodel, **kwds)

    A stochastic variable valued as a Gaussian process realization.

    :Arguments:

        -   `name`: The name of the variable.

        -   `submodel`: The Gaussian process submodel to which the
            variable belongs.

        -   `relative_precision`: See documentation.

    :SeeAlso: GPSubmodel, Realization, GPEvaluation
    """

    def __init__(self, name, submodel, trace=True, value=None, rseed=False,
                 observed=False, cache_depth=2, plot=None, verbose=-1):

        self.submodel = submodel

        pm.Stochastic.__init__(self, gp_logp, GaussianProcess.__doc__, name,
                               {'M': submodel.M,
                                'C': submodel.C,
                                'mesh': submodel.mesh,
                                'f_eval': submodel.f_eval,
                                'M_obs': submodel.M_obs,
                                'C_obs': submodel.C_obs},
                               gp_rand, trace, value, np.dtype('object'), rseed, observed, cache_depth, plot, verbose, False)

        self.rand()

    def gen_lazy_function(self):
        pass

    def get_logp(self):
        raise TypeError('Gaussian process %s has no logp.' % self.__name__)

    def set_logp(self, new_logp):
        raise TypeError('Gaussian process %s has no logp.' % self.__name__)

    def set_value(self, new_value):
        # If this is a new value pulled off the trace, save it the trouble of
        # observing
        if np.all(new_value.x_sofar == pm.utils.value(self.parents['mesh'])):
            if new_value.need_init_obs:
                new_value.M_internal = copy.copy(
                    pm.utils.value(
                        self.parents[
                            'M_obs']))
                new_value.C_internal = copy.copy(
                    pm.utils.value(
                        self.parents[
                            'C_obs']))
                new_value.need_init_obs = False
        pm.Stochastic.set_value(self, new_value)

    logp = property(fget=get_logp, fset=set_logp)

# for attr in ['neg','abs','invert'] + ['add', 'mul', 'sub'] + ['iadd','isub','imul','idiv','itruediv','ifloordiv','imod','ipow','ilshift','irshift','iand','ixor','ior','unicode']:
#     GaussianProcess.__delattr__('__%s__'%attr)


class GPEvaluation(pm.MvNormalChol):
    pass


class GPSubmodel(pm.ObjectContainer):

    """
    G=GaussianProcess(name, M, C, mesh[, init_vals, obs_on_mesh, tally_all])

    A Gaussian process submodel, which is a container for GaussianProcess and
    GPEvaluation objects.

    :Arguments:

        -   `name`: A prefix for the names of all the variables.

        -   `M`, `C`: The mean and covariance serving as parents of
            the submodel.

        -   `mesh`: See documentation.

        -   `init_vals`: The initial value of self.f_eval

        -   `obs_on_mesh`: Whether self.f_eval's value is observed.

        -   `tally_all`: Whether all variables should be tallied, or just
                f and f_eval.

    :Attributes:

        -   `f`: A GaussianProcess object, which is valued as a GP realization
            with prior mean self.M and prior covariance self.C.

        -   `f_eval`: The evaluation of self.f on self.mesh. This is a
            multivariate normal variable with a logp attribute.

        -   `mesh`: The mesh that was input, wrapped in a deterministic.

        -   `C_eval`: The evaluation of self.C on self.mesh.

        -   `S_eval`: The lower-triangular Cholesky factor of C_eval.

        -   `M_obs`, `C_obs`: Versions of self.M and self.C, observed on
                self.mesh with value self.f_eval.value.

        -   `fr_check`: A potential that enforces the constraint that
                self.C_eval must be full-rank.


    :SeeAlso: GaussianProcess, Realization, GPEvaluation
    """

    def __init__(self, name, M, C, mesh, init_vals=None,
                 obs_on_mesh=False, tally_all=False, **kwds):

        if isinstance(mesh, pm.Variable):
            mesh = pm.Lambda(
                '%s_mesh' %
                name,
                lambda mesh=mesh: pm.gp.regularize_array(
                    mesh),
                trace=False)
        else:
            mesh = pm.gp.regularize_array(mesh)
        self.name = name

        self.mesh = mesh
        self.M = M
        self.C = C

        if len(mesh) == 0:
            self.M_obs = M
            self.C_obs = C
            self.f_eval = None
        else:

            @pm.deterministic(
                trace=tally_all, name='%s_covariance_bits' % name)
            def covariance_bits(C=C, mesh=mesh):
                """
                Both the realization 'f' and the on-mesh evaluation 'f_eval' need the
                Cholesky factor of the covariance evaluation. The Gibbs step method
                also needs the full covariance evaluation. The mean needs a certain other
                function of the full covariance evaluation.

                All these things can be got as byproducts of Covariance.observe. Keeping the
                observed covariance and using it as the parent of f means the computations only
                get done once.
                """
                C_obs = copy.copy(C)
                try:
                    U, C_eval, Uo_Cxo = C_obs.observe(
                        mesh, np.zeros(mesh.shape[0]), output_type='s')
                    return U.T.copy('F'), C_eval, C_obs, Uo_Cxo
                except np.linalg.LinAlgError:
                    return None

            S_eval = pm.Lambda(
                '%s_S_eval' %
                name,
                lambda cb=covariance_bits: cb[
                    0] if cb else None,
                doc="The lower triangular Cholesky factor of %s.C_eval" %
                name,
                trace=tally_all or kwds.get(
                    'tally_S_eval',
                    False))
            C_eval = pm.Lambda(
                '%s_C_eval' % name,
                lambda cb=covariance_bits: cb[
                    1] if cb else None,
                doc="The evaluation %s.C(%s.mesh, %s.mesh)" % (
                    name,
                    name,
                    name),
                trace=tally_all or kwds.get(
                    'tally_C_eval',
                    False))
            C_obs = pm.Lambda(
                '%s_C_obs' % name,
                lambda cb=covariance_bits: cb[
                    2] if cb else None,
                doc="%s.C, observed on %s.mesh" % (
                    name,
                    name),
                trace=tally_all or kwds.get(
                    'tally_C_obs',
                    False))
            Uo_Cxo = pm.Lambda(
                '%s_Uo_Cxo' % name,
                lambda cb=covariance_bits: cb[
                    3] if cb else None,
                doc="A byproduct of observation of %s.C that can be used by %s.M" % (
                    name,
                    name),
                trace=tally_all or kwds.get(
                    'tally_Uo_Cxo',
                    False))
            M_eval = pm.Lambda(
                '%s_M_eval' % name,
                lambda M=M,
                mesh=mesh,
                Uo_Cxo=Uo_Cxo: M(
                    mesh,
                    Uo_Cxo=Uo_Cxo),
                trace=tally_all or kwds.get(
                    'tally_M_eval',
                    False),
                doc="The evaluation %s.M(%s.mesh)" % (
                    name,
                    name))

            @pm.potential(name='%s_fr_check' % name)
            def fr_check(S_eval=S_eval):
                """
                Forbids non-positive-definite C_evals.
                """
                if S_eval is None:
                    return -np.inf
                else:
                    return 0
            fr_check = fr_check

            f_eval = GPEvaluation(
                '%s_f_eval' % name, mu=M_eval, sig=S_eval, value=init_vals, trace=kwds.get('tally_f_eval', True), observed=obs_on_mesh,
                doc="The evaluation %s.f(%s.mesh).\nThis is a multivariate normal variable with mean %s.M_eval and covariance %s.C_eval." % (name, name, name, name))

            @pm.deterministic(
                trace=tally_all or kwds.get('tally_M_obs', False), name='%s_M_obs' % name)
            def M_obs(M=M, f_eval=f_eval, C_obs=C_obs, mesh=mesh):
                """
                Creates an observed mean object to match %sC_obs.
                """ % name
                M_obs = copy.copy(M)
                M_obs.observe(C_obs, mesh, f_eval)
                return M_obs

            self.M_obs = M_obs
            self.C_obs = C_obs
            self.f_eval = f_eval

        f = GaussianProcess(
            '%s_f' %
            name,
            self,
            trace=tally_all or kwds.get(
                'tally_f',
                True))

        f.rand()
        l = locals()
        lk = filter(lambda k: isinstance(l[k], pm.Node), l.keys())
        l = dict([(k, l[k]) for k in lk])
        pm.ObjectContainer.__init__(self, l)

    def getobjects(self):
        names = ['M_eval', 'C_eval', 'S_eval', 'f_eval', 'f', 'fr_check']
        return dict(zip(['%s_%s' % (self.name, name) for name in names], [
                    getattr(self, name) for name in names]))



#!/usr/bin/env python
"""
zip.py

Zero-inflated Poisson example using simulated data.
"""
import numpy as np
from pymc import Uniform, Beta, observed, rpoisson, poisson_like

# True parameter values
mu_true = 5
psi_true = 0.75
n = 100

# Simulate some data
data = np.array([rpoisson(mu_true) * (np.random.random() < psi_true)
                 for i in range(n)])

# Uniorm prior on Poisson mean
mu = Uniform('mu', 0, 20)
# Beta prior on psi
psi = Beta('psi', alpha=1, beta=1)


@observed(dtype=int, plot=False)
def zip(value=data, mu=mu, psi=psi):
    """ Zero-inflated Poisson likelihood """

    # Initialize likeihood
    like = 0.0

    # Loop over data
    for x in value:

        if not x:
            # Zero values
            like += np.log((1. - psi) + psi * np.exp(-mu))

        else:
            # Non-zero values
            like += np.log(psi) + poisson_like(x, mu)

    return like

if __name__ == "__main__":

    from pymc import MCMC, Matplot

    # Run model and plot posteriors
    M = MCMC(locals())
    M.sample(100000, 50000)
    Matplot.plot(M)



"""
pymc.CommonDeterministics

A collection of Deterministic subclasses to handle common situations.
It's a good idea to use these rather than user-defined objects when
possible, as some fitting methods (particularly Gibbs step methods)
will know how to handle them but not user-defined objects with
equivalent functionality.
"""

__docformat__ = 'reStructuredText'

from . import PyMCObjects as pm
from .Node import Variable
from .Container import Container
from .InstantiationDecorators import deterministic, check_special_methods
import numpy as np
from numpy import sum, shape, size, ravel, sign, zeros, ones, broadcast, newaxis
import inspect
import types
from .utils import safe_len, stukel_logit, stukel_invlogit, logit, invlogit, value, find_element, get_signature
from copy import copy
import sys
import operator
try:
    import builtins    # Python 3
except ImportError:
    import __builtin__ as builtins  # Python 2
try:
    from types import UnboundMethodType
except ImportError:
    # On Python 3, unbound methods are just functions.
    def UnboundMethodType(func, inst, cls):
        return func

from . import six
xrange = six.moves.xrange

__all__ = [
    'CompletedDirichlet', 'LinearCombination', 'Index', 'Lambda', 'lambda_deterministic', 'lam_dtrm',
    'logit', 'invlogit', 'stukel_logit', 'stukel_invlogit', 'Logit', 'InvLogit', 'StukelLogit', 'StukelInvLogit',
    'pfunc']  # +['iter_','complex_','int_','long_','float_','oct_','hex_']


class Lambda(pm.Deterministic):

    """
    L = Lambda(name, lambda p1=p1, p2=p2: f(p1, p2)[,
        doc, dtype=None, trace=True, cache_depth=2, plot=None])

    Converts second argument, an anonymous function, into a
    Deterministic object with specified name.

    :Parameters:
      name : string
        The name of the deteriministic object to be created.
      lambda : function
        The function from which the deterministic object should
        be created. All arguments must be given default values!
      p1, p2, ... : any
        The parameters of lambda.
      other parameters :
        See docstring of Deterministic.

    :Note:
      Will work even if argument 'lambda' is a named function
      (defined using def)

    :SeeAlso:
      Deterministic, Logit, StukelLogit, StukelInvLogit, Logit, InvLogit,
      LinearCombination, Index
    """

    def __init__(self, name, lam_fun,
                 doc='A Deterministic made from an anonymous function', *args, **kwds):
        (parent_names, parent_values) = get_signature(lam_fun)

        if parent_values is None:
            raise ValueError(
                '%s: All arguments to lam_fun must have default values.' %
                name)

        if not len(parent_names) == len(parent_values):
            raise ValueError(
                '%s: All arguments to lam_fun must have default values.' %
                name)

        parents = dict(zip(parent_names[-len(parent_values):], parent_values))

        pm.Deterministic.__init__(
            self,
            eval=lam_fun,
            name=name,
            parents=parents,
            doc=doc,
            *args,
            **kwds)


def lambda_deterministic(*args, **kwargs):
    """
    An alias for Lambda

    :SeeAlso:
      Lambda
    """
    return Lambda(*args, **kwargs)


def lam_dtrm(*args, **kwargs):
    """
    An alias for Lambda

    :SeeAlso:
      Lambda
    """
    return Lambda(*args, **kwargs)


class Logit(pm.Deterministic):

    """
    L = Logit(name, theta[, doc, dtype=None, trace=True,
        cache_depth=2, plot=None])

    A deterministic variable whose value is the logit of parent theta.

    :Parameters:
      name : string
        The name of the variable.
      theta : number, array or variable
        The parent to which the logit function should be applied.
        Must be between 0 and 1.
      other parameters :
        See docstring of Deterministic.

    :SeeAlso:
      Deterministic, Lambda, InvLogit, StukelLogit, StukelInvLogit
    """

    def __init__(self, name, theta,
                 doc='A logit transformation', *args, **kwds):
        pm.Deterministic.__init__(
            self,
            eval=logit,
            name=name,
            parents={'theta': theta},
            doc=doc,
            *args,
            **kwds)


class InvLogit(pm.Deterministic):

    """
    P = InvLogit(name, ltheta[, doc, dtype=None, trace=True,
        cache_depth=2, plot=None])

    A Deterministic whose value is the inverse logit of parent ltheta.

    :Parameters:
      name : string
        The name of the variable.
      ltheta : number, array or variable
        The parent to which the inverse logit function should be
        applied.
      other parameters :
        See docstring of Deterministic.

    :SeeAlso:
      Deterministic, Lambda, Logit, StukelLogit, StukelInvLogit
    """

    def __init__(self, name, ltheta,
                 doc='An inverse logit transformation', *args, **kwds):
        pm.Deterministic.__init__(
            self,
            eval=invlogit,
            name=name,
            parents={
                'ltheta': ltheta},
            doc=doc,
            *args,
            **kwds)


class StukelLogit(pm.Deterministic):

    """
    S = StukelLogit(name, theta, a1, a2, [, doc, dtype=None, trace=True,
        cache_depth=2, plot=None])

    A Deterministic whose value is Stukel's link function with
    parameters a1 and a2 applied to theta.

    To see the effects of a1 and a2, try plotting the function stukel_logit
    on theta=linspace(.1,.9,100)

    :Parameters:
      name : string
        The name of the variable.
      theta : number, array or variable.
        The parent to which the link function should be
        applied. Must be between 0 and 1.
      a1 : number
        One of the shape parameters.
      a2 : number
        The other shape parameter.
      other parameters :
        See docstring of Deterministic.

    :Reference:
      Therese A. Stukel, 'Generalized Logistic Models',
      JASA vol 83 no 402, pp.426-431 (June 1988)

    :SeeAlso:
      Deterministic, Lambda, Logit, InvLogit, StukelInvLogit
    """

    def __init__(self, name, theta, a1, a2,
                 doc="Stukel's link function", *args, **kwds):
        pm.Deterministic.__init__(self, eval=stukel_logit,
                                  name=name, parents={
                                      'theta': theta, 'a1': a1, 'a2': a2},
                                  doc=doc, *args, **kwds)


class StukelInvLogit(pm.Deterministic):

    """
    P = StukelInvLogit(name, ltheta, a1, a2, [, doc, dtype=None,
        trace=True, cache_depth=2, plot=None])

    A Deterministic whose value is Stukel's inverse link function with
    parameters a1 and a2 applied to ltheta.

    To see the effects of a1 and a2, try plotting the function stukel_invlogit
    on ltheta=linspace(-5,5,100)

    :Parameters:
      name : string
        The name of the variable.
      ltheta : number, array or variable.
        The parent to which the inverse link function should
        be applied. Must be between 0 and 1.
      a1 : number
        One of the shape parameters.
      a2 : number
        The other shape parameter.
      other parameters :
        See docstring of Deterministic.

    :Reference:
      Therese A. Stukel, 'Generalized Logistic Models',
      JASA vol 83 no 402, pp.426-431 (June 1988)

    :SeeAlso:
      Deterministic, Lambda, Logit, InvLogit, StukelLogit
    """

    def __init__(self, name, ltheta, a1, a2,
                 doc="Stukel's inverse link function", *args, **kwds):
        pm.Deterministic.__init__(self, eval=stukel_invlogit,
                                  name=name, parents={
                                      'ltheta': ltheta, 'a1': a1, 'a2': a2},
                                  doc=doc, *args, **kwds)


class CompletedDirichlet(pm.Deterministic):

    """
    CD = CompletedDirichlet(name, D[, doc, trace=True,
        cache_depth=2, plot=None])

    'Completes' the value of D by appending 1-sum(D.value) to the end.

    :Parameters:
      name : string
        The name of the variable.
      D : array or variable
        Value of object will be 1-sum(D) or 1-sum(D.value).
        Sum of D or D's value must be between 0 and 1.
      other parameters:
        See docstring of Deterministic

    :SeeAlso:
      Deterministic, Lambda, Index, LinearCombination
    """

    def __init__(self, name, D, doc=None, trace=True,
                 cache_depth=2, plot=None, verbose=-1):

        def eval_fun(D):
            N = len(D)
            out = np.empty((1, N + 1))
            out[0, :N] = D
            out[0, N] = 1. - np.sum(D)
            return out

        if doc is None:
            doc = 'The completed version of %s' % D.__name__

        pm.Deterministic.__init__(
            self, eval=eval_fun, name=name, parents={'D': D}, doc=doc,
            dtype=float, trace=trace, cache_depth=cache_depth, plot=plot, verbose=verbose)


class LinearCombination(pm.Deterministic):

    """
    L = LinearCombination(name, x, y[, doc, dtype=None,
        trace=True, cache_depth=2, plot=None])

    A Deterministic returning the sum of dot(x[i],y[i]).

    :Parameters:
      name : string
        The name of the variable
      x : list or variable
        Will be multiplied against y and summed.
      y : list or variable
        Will be multiplied against x and summed.
      other parameters :
        See docstring of Deterministic.

    :Attributes:
      x : list or variable
        Input argument
      y : list or variable
        Input argument
      N : integer
        length of x and y
      coefs : dictionary
        Keyed by variable. Indicates what each variable is multiplied by.
      sides : dictionary
        Keyed by variable. Indicates whether each variable is in x or y.
      offsets : dictionary
        Keyed by variable. Indicates everything that gets added to each
        stochastic and its coefficient.

    :SeeAlso:
      Deterministic, Lambda, Index
    """

    def __init__(self, name, x, y,
                 doc='A linear combination of several variables', *args, **kwds):
        self.x = x
        self.y = y
        self.N = len(self.x)

        if not len(self.y) == len(self.x):
            raise ValueError('Arguments x and y must be same length.')

        def eval_fun(x, y):
            out = np.dot(x[0], y[0])
            for i in xrange(1, len(x)):
                out = out + np.dot(x[i], y[i])
            return np.asarray(out).squeeze()

        pm.Deterministic.__init__(self,
                                  eval=eval_fun,
                                  doc=doc,
                                  name=name,
                                  parents={'x': x, 'y': y},
                                  *args, **kwds)

        # Tabulate coefficients and offsets of each constituent Stochastic.
        self.coefs = {}
        self.sides = {}

        for s in self.parents.stochastics | self.parents.observed_stochastics:
            self.coefs[s] = []
            self.sides[s] = []

        for i in xrange(self.N):

            stochastic_elem = None

            if isinstance(x[i], pm.Stochastic):

                if x[i] is y[i]:
                    raise ValueError(
                        'Stochastic %s multiplied by itself in LinearCombination %s.' %
                        (x[i], self))

                stochastic_elem = x[i]
                self.sides[stochastic_elem].append('L')
                this_coef = Lambda(
                    '%s_coef' %
                    stochastic_elem,
                    lambda c=y[
                        i]: np.asarray(
                            c))
                self.coefs[stochastic_elem].append(this_coef)

            if isinstance(y[i], pm.Stochastic):

                stochastic_elem = y[i]
                self.sides[stochastic_elem].append('R')
                this_coef = Lambda(
                    '%s_coef' %
                    stochastic_elem,
                    lambda c=x[
                        i]: np.asarray(
                            c))
                self.coefs[stochastic_elem].append(this_coef)

        self.sides = Container(self.sides)
        self.coefs = Container(self.coefs)

# TODO: Index should be special-cased in the future.
# TODO: - It should be a subclass of LinearCombination.
# TODO:   Reason: The Gibbs samplers should be able to recognize it as a linear combination.
# TODO: - It should be considered an 'ultimate argument' by LazyFunction, so that it is checked for changes rather
# TODO:   than its parents.
# TODO:   Reason: If parents change at elements that aren't selected, here's no point having all the descendants
# TODO:   recompute.


class Index(pm.Deterministic):

    """
    I = Index(name, x, index[, doc, dtype=None, trace=True,
        cache_depth=2, plot=None])

    A deterministic returning x[index].

    Useful for hierarchical models/ clustering/ discriminant analysis.
    Emulates LinearCombination to make it easier to write Gibbs step
    methods that can deal with such cases.

    :Parameters:
      name : string
        The name of the variable
      x : list or variable
        Will be indexed.
      index : integer or variable
        Index to use when computing value.
      other parameters :
        See docstring of Deterministic.

    :Attributes:
      index : variable
        Valued as current index.
      x:
        Variable that gets sliced.

    :SeeAlso:
      Deterministic, Lambda, LinearCombination
    """

    def __init__(self, name, x, index,
                 doc="Selects one of a list of several variables", *args, **kwds):
        self.index = Lambda('index', lambda i=index: np.int(i))
        self.x = x

        def eval_fun(x, index):
            return x[index]

        pm.Deterministic.__init__(self,
                                  eval=eval_fun,
                                  doc=doc,
                                  name='%s[%s]' % (str(x), str(index)),
                                  parents={'x': x, 'index': self.index},
                                  *args, **kwds)

# =================================================================
# = pfunc converts ordinary functions to Deterministic factories. =
# =================================================================


def pufunc(func):
    """
    Called by pfunc to convert NumPy ufuncs to deterministic factories.
    """
    def dtrm_generator(*args):
        if len(args) != func.nin:
            raise ValueError('invalid number of arguments')
        name = func.__name__ + '(' + '_'.join(
            [str(arg) for arg in list(args)]) + ')'
        doc_str = 'A deterministic returning %s(%s)' % (
            func.__name__,
            ', '.join([str(arg) for arg in args]))
        parents = {}
        for i in xrange(func.nin):
            parents['in%i' % i] = args[i]

        def wrapper(**kwargs):
            return func(*[kwargs['in%i' % i] for i in xrange(func.nin)])
        return pm.Deterministic(
            wrapper, doc_str, name, parents, trace=False, plot=False)
    dtrm_generator.__name__ = func.__name__ + '_deterministic_generator'
    dtrm_generator.__doc__ = """
Deterministic-generating wrapper for %s. Original docstring:
%s

%s
    """ % (func.__name__, '_' * 60, func.__doc__)
    return dtrm_generator


def pfunc(func):
    """
    pf = pfunc(func)

    Returns a function that can be called just like func; however its arguments may be
    PyMC objects or containers of PyMC objects, and its return value will be a deterministic.

    Example:

        >>> A = pymc.Normal('A',0,1,size=10)
        >>> pprod = pymc.pfunc(numpy.prod)
        >>> B = pprod(A, axis=0)
        >>> B
        <pymc.PyMCObjects.Deterministic 'prod(A_0)' at 0x3ce49b0>
        >>> B.value
        -0.0049333289649554912
        >>> numpy.prod(A.value)
        -0.0049333289649554912
    """
    if isinstance(func, np.ufunc):
        return pufunc(func)
    elif not inspect.isfunction(func):
        if func.__name__ == '__call__':
            raise ValueError(
                'Cannot get argspec of call method. Is it builtin?')
        try:
            return pfunc(func.__call__)
        except:
            cls, inst, tb = sys.exc_info()
            inst = cls(
                'Failed to create pfunc wrapper from object %s. Original error message:\n\n%s' %
                (func, inst.message))
            six.reraise(cls, inst, tb)
    fargs, fdefaults = get_signature(func)
    n_fargs = len(fargs)

    def dtrm_generator(*args, **kwds):
        name = func.__name__ + '(' + '_'.join([str(arg)
                                               for arg in list(args) +
                                               list(kwds.values())]) + ')'
        doc_str = 'A deterministic returning %s(%s, %s)' % (
            func.__name__,
            ', '.join([str(arg) for arg in args]),
            ', '.join(['%s=%s' % (key,
                                  str(val)) for key,
                       val in six.iteritems(kwds)]))

        parents = {}
        varargs = []
        for kwd, val in six.iteritems(kwds):
            parents[kwd] = val
        for i in xrange(len(args)):
            if i < n_fargs:
                parents[fargs[i]] = args[i]
            else:
                varargs.append(args[i])

        if len(varargs) == 0:
            eval_fun = func
        else:
            parents['varargs'] = varargs

            def wrapper(**wkwds_in):
                wkwds = copy(wkwds_in)
                wargs = []
                for arg in fargs:
                    wargs.append(wkwds.pop(arg))
                wargs.extend(wkwds.pop('varargs'))
                return func(*wargs, **wkwds)
            eval_fun = wrapper

        return pm.Deterministic(
            eval_fun, doc_str, name, parents, trace=False, plot=False)
    dtrm_generator.__name__ = func.__name__ + '_deterministic_generator'
    dtrm_generator.__doc__ = """
Deterministic-generating wrapper for %s. Original docstring:
%s

%s
    """ % (func.__name__, '_' * 60, func.__doc__)
    return dtrm_generator


# ==========================================================
# = Add special methods to variables to support FBC syntax =
# ==========================================================

def create_uni_method(op_name, klass, jacobians=None):
    """
    Creates a new univariate special method, such as A.__neg__() <=> -A,
    for target class. The method is called __op_name__.
    """
    # This function will become the actual method.
    op_modules = [operator, builtins]
    op_names = [op_name, op_name + '_']

    op_function_base = find_element(op_names, op_modules, error_on_fail=True)
    # many such functions do not take keyword arguments, so we need to wrap
    # them

    def op_function(self):
        return op_function_base(self)

    def new_method(self):
        # This code creates a Deterministic object.
        if not check_special_methods():
            raise NotImplementedError(
                'Special method %s called on %s, but special methods have been disabled. Set pymc.special_methods_available to True to enable them.' %
                (op_name, str(self)))

        jacobian_formats = {'self': 'transformation_operation'}
        return pm.Deterministic(op_function,
                                'A Deterministic returning the value of %s(%s)' % (
                                    op_name, self.__name__),
                                '(' + op_name + '_' + self.__name__ + ')',
                                parents={'self': self},
                                trace=False,
                                plot=False,
                                jacobians=jacobians,
                                jacobian_formats=jacobian_formats)
    # Make the function into a method for klass.

    new_method.__name__ = '__' + op_name + '__'
    setattr(
        klass,
        new_method.__name__,
        UnboundMethodType(
            new_method,
            None,
            klass))


def create_casting_method(op, klass):
    """
    Creates a new univariate special method, such as A.__float__() <=> float(A.value),
    for target class. The method is called __op_name__.
    """
    # This function will become the actual method.

    def new_method(self, op=op):
        if not check_special_methods():
            raise NotImplementedError(
                'Special method %s called on %s, but special methods have been disabled. Set pymc.special_methods_available to True to enable them.' %
                (op_name, str(self)))
        return op(self.value)
    # Make the function into a method for klass.
    new_method.__name__ = '__' + op.__name__ + '__'
    setattr(
        klass,
        new_method.__name__,
        UnboundMethodType(
            new_method,
            None,
            klass))


def create_rl_bin_method(op_name, klass, jacobians={}):
    """
    Creates a new binary special method with left and right versions, such as
        A.__mul__(B) <=> A*B,
        A.__rmul__(B) <=> [B*A if B.__mul__(A) fails]
    for target class. The method is called __op_name__.
    """
    # Make left and right versions.
    for prefix in ['r', '']:
        # This function will became the methods.
        op_modules = [operator, builtins]
        op_names = [op_name, op_name + '_']

        op_function_base = find_element(
            op_names,
            op_modules,
            error_on_fail=True)
        # many such functions do not take keyword arguments, so we need to wrap
        # them

        def op_function(a, b):
            return op_function_base(a, b)

        def new_method(self, other, prefix=prefix):
            if not check_special_methods():
                raise NotImplementedError(
                    'Special method %s called on %s, but special methods have been disabled. Set pymc.special_methods_available to True to enable them.' %
                    (op_name, str(self)))
            # This code will create one of two Deterministic objects.
            if prefix == 'r':
                parents = {'a': other, 'b': self}

            else:
                parents = {'a': self, 'b': other}
            jacobian_formats = {'a': 'broadcast_operation',
                                'b': 'broadcast_operation'}
            return pm.Deterministic(op_function,
                                    'A Deterministic returning the value of %s(%s,%s)' % (
                                        prefix + op_name, self.__name__, str(
                                            other)),
                                    '(' + '_'.join(
                                        [self.__name__,
                                         prefix + op_name,
                                         str(other)]) + ')',
                                    parents,
                                    trace=False,
                                    plot=False,
                                    jacobians=jacobians,
                                    jacobian_formats=jacobian_formats)
        # Convert the functions into methods for klass.
        new_method.__name__ = '__' + prefix + op_name + '__'
        setattr(
            klass,
            new_method.__name__,
            UnboundMethodType(
                new_method,
                None,
                klass))


def create_rl_lin_comb_method(op_name, klass, x_roles, y_roles):
    """
    Creates a new binary special method with left and right versions, such as
        A.__mul__(B) <=> A*B,
        A.__rmul__(B) <=> [B*A if B.__mul__(A) fails]
    for target class. The method is called __op_name__.
    """
    # This function will became the methods.

    def new_method(self, other, x_roles=x_roles, y_roles=y_roles):
        if not check_special_methods():
            raise NotImplementedError(
                'Special method %s called on %s, but special methods have been disabled. Set pymc.special_methods_available to True to enable them.' %
                (op_name, str(self)))
        x = []
        y = []
        for xr in x_roles:
            if xr == 'self':
                x.append(self)
            elif xr == 'other':
                x.append(other)
            else:
                x.append(xr)
        for yr in y_roles:
            if yr == 'self':
                y.append(self)
            elif yr == 'other':
                y.append(other)
            else:
                y.append(yr)
        # This code will create one of two Deterministic objects.
        return LinearCombination(
            '(' + '_'.join([self.__name__, op_name, str(other)]) + ')', x, y, trace=False, plot=False)

    # Convert the functions into methods for klass.
    new_method.__name__ = '__' + op_name + '__'
    setattr(
        klass,
        new_method.__name__,
        UnboundMethodType(
            new_method,
            None,
            klass))


def create_bin_method(op_name, klass):
    """
    Creates a new binary special method with only a left version, such as
    A.__eq__(B) <=> A==B, for target class. The method is called __op_name__.
    """
    # This function will become the method.

    def new_method(self, other):
        if not check_special_methods():
            raise NotImplementedError(
                'Special method %s called on %s, but special methods have been disabled. Set pymc.special_methods_available to True to enable them.' %
                (op_name, str(self)))
        # This code creates a Deterministic object.

        def eval_fun(self, other, op):
            return getattr(self, op)(other)
        return pm.Deterministic(eval_fun,
                                'A Deterministic returning the value of %s(%s,%s)' % (
                                    op_name, self.__name__, str(other)),
                                '(' + '_'.join(
                                    [self.__name__,
                                     op_name,
                                     str(other)]) + ')',
                                {'self': self,
                                 'other': other,
                                 'op': '__' + op_name + '__'},
                                trace=False,
                                plot=False)
    # Convert the function into a method for klass.
    new_method.__name__ = '__' + op_name + '__'
    setattr(
        klass,
        new_method.__name__,
        UnboundMethodType(
            new_method,
            None,
            klass))


def create_nonimplemented_method(op_name, klass):
    """
    Creates a new method that raises NotImplementedError.
    """

    def new_method(self, *args):
        raise NotImplementedError(
            'Special method %s has not been implemented for PyMC variables.' %
            op_name)
    new_method.__name__ = '__' + op_name + '__'
    setattr(
        klass,
        new_method.__name__,
        UnboundMethodType(
            new_method,
            None,
            klass))


def op_to_jacobians(op, module):
    if isinstance(module, types.ModuleType):
        module = copy(module.__dict__)
    elif isinstance(module, dict):
        module = copy(module)
    else:
        raise AttributeError

    name = op + "_jacobians"
    try:
        jacobians = module[name]
    except:
        jacobians = {}

    return jacobians

# Left/right binary operators


truediv_jacobians = {'a': lambda a, b: ones(shape(a)) / b,
                     'b': lambda a, b: - a / b ** 2}

div_jacobians = truediv_jacobians

pow_jacobians = {'a': lambda a, b: b * a ** (b - 1.0),
                 'b': lambda a, b: np.log(a) * a ** b}


for op in ['truediv', 'floordiv', 'mod', 'divmod', 'pow', 'lshift', 'rshift', 'and', 'xor', 'or']:
    create_rl_bin_method(
        op,
        Variable,
        jacobians=op_to_jacobians(
            op,
            locals(
            )))

try:
    create_rl_bin_method(
        'div',
        Variable,
        jacobians=op_to_jacobians(
            'div',
            locals(
            )))
except NameError:
    pass    # Python 3 has only truediv and floordiv

# Binary operators eq not part of this set because it messes up having
# stochastics in lists
for op in ['lt', 'le', 'ne', 'gt', 'ge']:
    create_bin_method(op, Variable)


def equal(s1, s2):  # makes up for deficiency of __eq__
    return pm.Deterministic(lambda x1, x2: x1 == x2,
                            'A Deterministic returning the value of x1 == x2',
                            '(' + '_'.join([s1.__name__, '=', str(s2)]) + ')',
                            {'x1': s1, 'x2': s2},
                            trace=False,
                            plot=False)


# Unary operators
neg_jacobians = {'self': lambda self: -ones(shape(self))}

pos_jacobians = {'self': lambda self: np.ones(shape(self))}

abs_jacobians = {'self': lambda self: np.sign(self)}

# no need for pos and __index__ seems to cause a lot of problems
for op in ['neg', 'abs', 'invert']:
    create_uni_method(op, Variable, jacobians=op_to_jacobians(op, locals()))

# Casting operators
for op in [iter, complex, int, float, oct, hex]:
    create_casting_method(op, Variable)

try:
    create_casting_method(long, Variable)
except NameError:
    pass    # No long in Python 3

# Addition, subtraction, multiplication
# TODO: Uncomment once LinearCombination issues are ironed out.
# create_rl_lin_comb_method('add', Variable, ['self', 'other'], [1,1])
# create_rl_lin_comb_method('radd', Variable, ['self', 'other'], [1,1])
# create_rl_lin_comb_method('sub', Variable, ['self', 'other'], [1,-1])
# create_rl_lin_comb_method('rsub', Variable, ['self', 'other'], [-1,1])
# create_rl_lin_comb_method('mul', Variable, ['self'],['other'])
# create_rl_lin_comb_method('rmul', Variable, ['self'],['other'])

# TODO: Comment once LinearCombination issues are ironed out.

add_jacobians = {'a': lambda a, b: ones(broadcast(a, b).shape),
                 'b': lambda a, b: ones(broadcast(a, b).shape)}

mul_jacobians = {'a': lambda a, b: ones(shape(a)) * b,
                 'b': lambda a, b: ones(shape(b)) * a}

sub_jacobians = {'a': lambda a, b: ones(broadcast(a, b).shape),
                 'b': lambda a, b: -ones(broadcast(a, b).shape)}

for op in ['add', 'mul', 'sub']:
    create_rl_bin_method(
        op,
        Variable,
        jacobians=op_to_jacobians(
            op,
            locals(
            )))

for op in ['iadd', 'isub', 'imul', 'idiv', 'itruediv', 'ifloordiv', 'imod', 'ipow', 'ilshift', 'irshift', 'iand', 'ixor', 'ior']:
    create_nonimplemented_method(op, Variable)


def getitem_jacobian(self, index):
    return index


# Create __getitem__ method.
def __getitem__(self, index):
    if not check_special_methods():
        raise NotImplementedError(
            'Special method __index__ called on %s, but special methods have been disabled. Set pymc.special_methods_available to True to enable them.' %
            str(self))
    # If index is number or number-valued variable, make an Index object
    name = '%s[%s]' % (self.__name__, str(index))
    if np.isscalar(value(index)) and len(np.shape(self.value)) < 2:
        if np.isreal(value(index)):
            return Index(name, self, index, trace=False, plot=False)
    # Otherwise make a standard Deterministic.

    def eval_fun(self, index):
        return self[index]

    jacobians = {'self': getitem_jacobian}
    jacobian_formats = {'self': 'index_operation'}
    return pm.Deterministic(eval_fun,
                            'A Deterministic returning the value of %s[%s]' % (
                                self.__name__, str(index)),
                            name,
                            {'self': self, 'index': index},
                            trace=False,
                            plot=False,
                            jacobians=jacobians,
                            jacobian_formats=jacobian_formats)
Variable.__getitem__ = UnboundMethodType(__getitem__, None, Variable)

# Create __call__ method for Variable.


def __call__(self, *args, **kwargs):
    if not check_special_methods():
        raise NotImplementedError(
            'Special method __call__ called on %s, but special methods have been disabled. Set pymc.special_methods_available to True to enable them.' %
            str(self))

    def eval_fun(self, args=args, kwargs=kwargs):
        return self(*args, **kwargs)
    return pm.Deterministic(eval_fun,
                            'A Deterministic returning the value of %s(*%s, **%s)' % (
                                self.__name__, str(args), str(kwargs)),
                            self.__name__ + '(*%s, **%s)' % (
                                str(args), str(kwargs)),
                            {'self': self, 'args': args, 'kwargs': kwargs},
                            trace=False,
                            plot=False)
Variable.__call__ = UnboundMethodType(__call__, None, Variable)

# def __getitem__(self, index):
#     def eval_fun(self, index=index):
#         return self.__getitem__[index]
#     return pm.Deterministic(eval_fun,
#                             'A Deterministic returning the value of %s[%s]'%(self.__name__, str(index)),
#                             self.__name__+'[%s]'%str(index),
#                             {'self':self, 'index': index},
#                             trace=False,
#                             plot=False)
# Variable.__getitem__ = UnboundMethodType(__getitem__, None, Variable)


# These are not working
# nonworking_ops = ['iter','complex','int','long','float','oct','hex','coerce','contains','len']
# These should NOT be implemented because they are in-place updates.



"""HDF5 database module.

Store the traces in an HDF5 array using pytables.


Implementation Notes
--------------------

This version only supports numeric objects, and stores them in
extentable HDF5 arrays.  This allows the implementation to handle very
large data vectors.


Additional Dependencies
-----------------------
 * HDF5 version 1.6.5, required by pytables.
 * pytables version 2 and up.  <http://sourceforge.net/projects/pytables/>

"""

import os
import sys
import traceback
import warnings

import numpy as np
import pymc
import tables

from pymc.database import base, pickle
from pymc import six

__all__ = ['Trace', 'Database', 'load']

warn_tally = """
Error tallying %s, will not try to tally it again this chain.
Did you make all the same variables and step methods tallyable
as were tallyable last time you used the database file?

Error:

%s"""


#

class Trace(base.Trace):

    """HDF5 trace."""

    def tally(self, chain):
        """Adds current value to trace."""

        arr = np.asarray(self._getfunc())
        arr = arr.reshape((1,) + arr.shape)
        self.db._arrays[chain, self.name].append(arr)

    # def __getitem__(self, index):
    #     """Mimic NumPy indexing for arrays."""
    #     chain = self._chain
    #     if chain is not None:
    #         tables = [self.db._gettables()[chain],]
    #     else:
    #         tables = self.db._gettables()
    #     out = []
    #     for table in tables:
    #         out.append(table.col(self.name))
    #     if np.isscalar(chain):
    #         return out[0][index]
    #     else:
    #         return np.hstack(out)[index]
    def gettrace(self, burn=0, thin=1, chain=-1, slicing=None):
        """Return the trace (last by default).

        :Parameters:
        burn : integer
          The number of transient steps to skip.
        thin : integer
          Keep one in thin.
        chain : integer
          The index of the chain to fetch. If None, return all chains. The
          default is to return the last chain.
        slicing : slice object
          A slice overriding burn and thin assignement.
        """

        # XXX: handle chain == None case properly

        if chain is None:
            chain = -1
        chain = self.db.chains[chain]

        arr = self.db._arrays[chain, self.name]

        if slicing is not None:
            burn, stop, thin = slicing.start, slicing.stop, slicing.step

        if slicing is None or stop is None:
            stop = arr.nrows
            return np.asarray(arr.read(start=burn, stop=stop, step=thin))

    __call__ = gettrace

    # def length(self, chain=-1):
    #     """Return the length of the trace.

    #     :Parameters:
    #     chain : int or None
    #       The chain index. If None, returns the combined length of all chains.
    #     """
    #     if chain is not None:
    #         tables = [self.db._gettables()[chain],]
    #     else:
    #         tables = self.db._gettables()

    #     n = np.asarray([table.nrows for table in tables])
    #     return n.sum()


#

class Database(pickle.Database):

    """HDF5 database.

    Create an HDF5 file <model>.h5.  Each chain is stored in a group,
    and the stochastics and deterministics are stored as extendable
    arrays in each group.
    """

    def __init__(self, dbname, dbmode='a',
                 dbcomplevel=0, dbcomplib='zlib',
                 **kwds):
        """Create an HDF5 database instance, where samples are stored
        in extendable arrays.

        :Parameters:
        dbname : string
          Name of the hdf5 file.
        dbmode : {'a', 'w', 'r'}
          File mode: 'a': append, 'w': overwrite, 'r': read-only.
        dbcomplevel : integer (0-9)
          Compression level, 0: no compression.
        dbcomplib : string
          Compression library (zlib, bzip2, lzo)

        :Notes:
          * zlib has a good compression ratio, although somewhat slow, and
            reasonably fast decompression.
          * lzo is a fast compression library offering however a low compression
            ratio.
          * bzip2 has an excellent compression ratio but requires more CPU.
        """

        self.__name__ = 'hdf5ea'
        self.__Trace__ = Trace

        self.dbname = dbname
        self.mode = dbmode

        db_exists = os.path.exists(self.dbname)
        self._h5file = tables.open_file(self.dbname, self.mode)

        default_filter = tables.Filters(
            complevel=dbcomplevel,
            complib=dbcomplib)
        if self.mode == 'r' or (self.mode == 'a' and db_exists):
            self.filter = getattr(self._h5file, 'filters', default_filter)
        else:
            self.filter = default_filter

        self.trace_names = []
        self._traces = {}
        # self._states = {}
        self._chains = {}
        self._arrays = {}

        # load existing data
        existing_chains = [gr for gr in self._h5file.listNodes("/")
                           if gr._v_name[:5] == 'chain']

        for chain in existing_chains:
            nchain = int(chain._v_name[5:])
            self._chains[nchain] = chain

            names = []
            for array in chain._f_listNodes():
                name = array._v_name
                self._arrays[nchain, name] = array

                if name not in self._traces:
                    self._traces[name] = Trace(name, db=self)

                names.append(name)

            self.trace_names.append(names)

    @property
    def chains(self):
        return range(len(self._chains))

    @property
    def nchains(self):
        return len(self._chains)

    # def connect_model(self, model):
    #     """Link the Database to the Model instance.
    #     In case a new database is created from scratch, ``connect_model``
    #     creates Trace objects for all tallyable pymc objects defined in
    #     `model`.
    #     If the database is being loaded from an existing file, ``connect_model``
    #     restore the objects trace to their stored value.
    #     :Parameters:
    #     model : pymc.Model instance
    #       An instance holding the pymc objects defining a statistical
    #       model (stochastics, deterministics, data, ...)
    #     """
    # Changed this to allow non-Model models. -AP
    #     if isinstance(model, pymc.Model):
    #         self.model = model
    #     else:
    #         raise AttributeError('Not a Model instance.')
    # Restore the state of the Model from an existing Database.
    # The `load` method will have already created the Trace objects.
    #     if hasattr(self, '_state_'):
    #         names = set()
    #         for morenames in self.trace_names:
    #             names.update(morenames)
    #         for name, fun in six.iteritems(model._funs_to_tally):
    #             if name in self._traces:
    #                 self._traces[name]._getfunc = fun
    #                 names.remove(name)
    #         if len(names) > 0:
    #             raise RuntimeError("Some objects from the database"
    #                                + "have not been assigned a getfunc: %s"
    #                                   % ', '.join(names))
    def _initialize(self, funs_to_tally, length):
        """Create a group named ``chain#`` to store all data for this chain."""

        chain = self.nchains
        self._chains[chain] = self._h5file.create_group(
            '/', 'chain%d' % chain, 'chain #%d' % chain)

        for name, fun in six.iteritems(funs_to_tally):

            arr = np.asarray(fun())

            assert arr.dtype != np.dtype('object')

            array = self._h5file.createEArray(
                self._chains[chain], name,
                tables.Atom.from_dtype(arr.dtype), (0,) + arr.shape,
                filters=self.filter)

            self._arrays[chain, name] = array
            self._traces[name] = Trace(name, getfunc=fun, db=self)
            self._traces[name]._initialize(self.chains, length)

        self.trace_names.append(list(funs_to_tally.keys()))

    def tally(self, chain=-1):
        chain = self.chains[chain]
        for name in self.trace_names[chain]:
            try:
                self._traces[name].tally(chain)
                self._arrays[chain, name].flush()
            except:
                cls, inst, tb = sys.exc_info()
                warnings.warn(warn_tally
                              % (name, ''.join(traceback.format_exception(cls, inst, tb))))
                self.trace_names[chain].remove(name)

    # def savestate(self, state, chain=-1):
    #     """Store a dictionnary containing the state of the Model and its
    #     StepMethods."""
    #     chain = self.chains[chain]
    #     if chain in self._states:
    #         self._states[chain] = state
    #     else:
    #         s = self._h5file.create_vlarray(chain,'_state_',tables.ObjectAtom(),title='The saved state of the sampler',filters=self.filter)
    #         s.append(state)
    #     self._h5file.flush()
    # def getstate(self, chain=-1):
    #     if len(self._chains)==0:
    #         return {}
    #     elif hasattr(self._chains[chain],'_state_'):
    #         if len(self._chains[chain]._state_)>0:
    #             return self._chains[chain]._state_[0]
    #         else:
    #             return {}
    #     else:
    #         return {}
    def _finalize(self, chain=-1):
        self._h5file.flush()

    def close(self):
        self._h5file.close()


def load(dbname, dbmode='a'):
    """Load an existing hdf5 database.

    Return a Database instance.

    :Parameters:
      filename : string
        Name of the hdf5 database to open.
      mode : 'a', 'r'
        File mode : 'a': append, 'r': read-only.
    """
    if dbmode == 'w':
        raise AttributeError("dbmode='w' not allowed for load.")
    db = Database(dbname, dbmode=dbmode)

    return db



"""
RAM database module

Store the trace in memory using NumPy arrays.

Implementation Notes
--------------------
This is the only backend using preallocated memory. All others simply
append values to a stack. It might be worthwhile to use a list instead
of a NumPy array to 1. simplify this backend, 2. standardize the
`Trace model` and 3. remove the need for a truncate method.
We would need to catch MemoryError exceptions though.
"""

import pymc
from numpy import zeros, shape, concatenate, ndarray, dtype
from . import base
import warnings
import numpy as np

__all__ = ['Trace', 'Database']


class Trace(base.Trace):

    """RAM Trace

    Store the samples in memory. No data is written to disk.
    """

    def __init__(self, name, getfunc=None, db=None, value=None):
        """Create a Trace instance.

        :Parameters:
        name : string
          The trace object name. This name should uniquely identify
          the pymc variable.
        getfunc : function
          A function returning the value to tally.
        db : Database instance
          The database owning this Trace.
        value : list
          The list of trace arrays. This is used when loading the Trace from
          disk."""
        if value is None:
            self._trace = {}
            self._index = {}
        else:
            self._trace = value
            self._index = dict(zip(value.keys(), list(map(len, value.values()))))

        base.Trace.__init__(self, name=name, getfunc=getfunc, db=db)

    def _initialize(self, chain, length):
        """Create an array of zeros with shape (length, shape(obj)), where
        obj is the internal PyMC Stochastic or Deterministic.
        """
        # If this db was loaded from the disk, it may not have its
        # tallied step methods' getfuncs yet.
        if self._getfunc is None:
            self._getfunc = self.db.model._funs_to_tally[self.name]

        # First, see if the object has an explicit dtype.
        value = np.array(self._getfunc())

        if value.dtype is object:
            self._trace[chain] = zeros(length, dtype=object)

        elif value.dtype is not None:
            self._trace[chain] = zeros((length,) + shape(value), value.dtype)

        # Otherwise, if it's an array, read off its value's dtype.
        elif isinstance(value, ndarray):
            self._trace[chain] = zeros((length,) + shape(value), value.dtype)

        # Otherwise, let numpy type its value. If the value is a scalar, the trace will be of the
        # corresponding type. Otherwise it'll be an object array.
        else:
            self._trace[chain] = zeros(
                (length,
                 ) + shape(value),
                dtype=value.__class__)

        self._index[chain] = 0

    def tally(self, chain):
        """Store the object's current value to a chain.

        :Parameters:
        chain : integer
          Chain index.
        """

        value = self._getfunc()

        try:
            self._trace[chain][self._index[chain]] = value.copy()
        except AttributeError:
            self._trace[chain][self._index[chain]] = value
        self._index[chain] += 1

    def truncate(self, index, chain):
        """
        Truncate the trace array to some index.

        :Parameters:
        index : int
          The index within the chain after which all values will be removed.
        chain : int
          The chain index (>=0).
        """
        self._trace[chain] = self._trace[chain][:index]

    def gettrace(self, burn=0, thin=1, chain=-1, slicing=None):
        """Return the trace.

        :Stochastics:
          - burn (int): The number of transient steps to skip.
          - thin (int): Keep one in thin.
          - chain (int): The index of the chain to fetch. If None, return all chains.
          - slicing: A slice, overriding burn and thin assignement.
        """
        if slicing is None:
            slicing = slice(burn, None, thin)
        if chain is not None:
            if chain < 0:
                chain = range(self.db.chains)[chain]
            return self._trace[chain][slicing]
        else:
            return concatenate(list(self._trace.values()))[slicing]

    def __getitem__(self, index):
        chain = self._chain
        if chain is None:
            return concatenate(list(self._trace.values()))[index]
        else:
            if chain < 0:
                chain = range(self.db.chains)[chain]
            return self._trace[chain][index]

    __call__ = gettrace

    def length(self, chain=-1):
        """Return the length of the trace.

        :Parameters:
        chain : int or None
          The chain index. If None, returns the combined length of all chains.
        """
        if chain is not None:
            if chain < 0:
                chain = range(self.db.chains)[chain]
            return self._trace[chain].shape[0]
        else:
            return sum([t.shape[0] for t in self._trace.values()])


class Database(base.Database):

    """RAM database.

    Store the samples in memory. No data is written to disk.
    """

    def __init__(self, dbname):
        """Create a RAM Database instance."""
        self.__name__ = 'ram'
        self.__Trace__ = Trace
        self.dbname = dbname
        self.trace_names = []
        # A list of sequences of names of the objects to tally.
        self._traces = {}  # A dictionary of the Trace objects.
        self.chains = 0



from numpy.testing import *
from pymc.gp import *
from .test_mean import M, x
from .test_cov import C
from numpy import *
from copy import copy

M = copy(M)
C = copy(C)

# Impose observations on the GP


class test_observation(TestCase):

    def test(self):

        obs_x = array([-.5, .5])
        V = array([.002, .002])
        data = array([3.1, 2.9])
        observe(M=M,
                C=C,
                obs_mesh=obs_x,
                obs_V=V,
                obs_vals=data)

        # Generate realizations

        for i in range(3):
            f = Realization(M, C)
            f(x)



# Copyright (c) Anand Patil, 2007

__docformat__ = 'reStructuredText'

from numpy import *
from .GPutils import regularize_array, trisolve

__all__ = ['Mean', 'zero_fn']


class Mean(object):

    """
    M = Mean(eval_fun, **params)


    A Gaussian process mean.


    :Arguments:

        -   `eval_fun`: A function that takes an argument x of shape  (n,ndim), where n is
            any integer and ndim is the dimensionality of  the space, or shape (n). In the
            latter case ndim should be assumed to be 1..

        -   `params`: Parameters to be passed to eval_fun


    :SeeAlso: Covariance, Realization, Observe
    """

    def __init__(self, eval_fun, **params):

        self.ndim = None
        self.observed = False
        self.obs_mesh = None
        self.base_mesh = None
        self.obs_vals = None
        self.obs_V = None
        self.reg_mat = None
        self.obs_len = None
        self.mean_under = None
        self.Uo = None
        self.obs_piv = None

        self.eval_fun = eval_fun
        self.params = params

    def observe(self, C, obs_mesh_new, obs_vals_new, mean_under=None):
        """
        Synchronizes self's observation status with C's.
        Values of observation are given by obs_vals.

        obs_mesh_new and obs_vals_new should already have
        been sliced, as Covariance.observe(..., output_type='o') does.
        """

        self.C = C
        self.obs_mesh = C.obs_mesh
        self.obs_len = C.obs_len

        self.Uo = C.Uo

        # Evaluate the underlying mean function on the new observation mesh.
        if mean_under is None:
            mean_under_new = C._mean_under_new(self, obs_mesh_new)
        else:
            mean_under_new = mean_under

        # If self hasn't been observed yet:
        if not self.observed:

            self.dev = (obs_vals_new - mean_under_new)

            self.reg_mat = C._unobs_reg(self)

        # If self has been observed already:
        elif len(obs_vals_new) > 0:

            # Rank of old observations.
            m_old = len(self.dev)

            # Deviation of new observation from mean without regard to old
            # observations.
            dev_new = (obs_vals_new - mean_under_new)

            # Again, basis covariances get special treatment.
            self.reg_mat = C._obs_reg(self, dev_new, m_old)

            # Stack deviations of old and new observations from unobserved
            # mean.
            self.dev = hstack((self.dev, dev_new))
        self.observed = True

    def __call__(self, x, observed=True, regularize=True, Uo_Cxo=None):

        # Record original shape of x and regularize it.
        orig_shape = shape(x)
        if len(orig_shape) > 1:
            orig_shape = orig_shape[:-1]

        if regularize:
            x = regularize_array(x)

        ndimx = x.shape[-1]
        lenx = x.shape[0]

        # Safety.
        if self.ndim is not None:
            if not self.ndim == ndimx:
                raise ValueError(
                    "The number of spatial dimensions of x does not match the number of spatial dimensions of the Mean instance's base mesh.")

        # Evaluate the unobserved mean
        M = self.eval_fun(x, **self.params).squeeze()

        # Condition. Basis covariances get special treatment. See documentation
        # for formulas.
        if self.observed and observed:
            M = self.C._obs_eval(self, M, x, Uo_Cxo)

        return M.reshape(orig_shape)


def zero_fn(x):
    "A function mapping any argument to zero."
    return zeros(x.shape[:-1])



# TODO: Make a better class for bases. Should be either a multidimensional generator or should output
# a multidimensional ndarray, and dots into the basis shouldn't require it to be a matrix.
# Note tensordot can do the dot products you need.

# Make an object-valued ndarray whose elements are index tuples.
# Make it a matrix whose elements are what you want, and then reshape it multidimensional, and see where the indices
# have to end up.

# Copyright (c) Anand Patil, 2007

__docformat__ = 'reStructuredText'

__all__ = ['BasisCovariance', 'SeparableBasisCovariance']

from numpy import *
from numpy.linalg import eigh, solve, cholesky, LinAlgError
from .GPutils import regularize_array, trisolve
from .linalg_utils import basis_diag_call
from .incomplete_chol import ichol_basis, ichol_full
from .Covariance import Covariance

from pymc import six
xrange = six.moves.xrange


class BasisCovariance(Covariance):

    """
    C=BasisCovariance(basis, coef_cov, relative_precision, **params)

    Realizations generated using such covariances will be of the form:
        R(x) = \sum basis(x).ravel()[i] * c[i],
        c ~ N(mu, coef_cov)

    :Arguments:

        -   `basis`:    An n-dimensional array of functions. Each should take arguments of shape (m, n),
                        where m may be any integer, followed by arbitrary keyword parameters.

        -   `coef_cov`: An array of shape basis.shape or basis.shape*2 (tuple multiplication). If
                        shape=basis.shape, the coefficients are assumed to be independent. Otherwise
                        coef_cov[i[1],...,i[n], j[1],...,j[n]] gives the prior covariance of coefficients
                        i[1],...,i[n] and j[1],...,j[n].

        -   `params`:   Parameters to be passed to basis.

        -   `relative_precision`: See documentation.


    :SeeAlso: Mean, Realization, Covariance, SeparableBasisCovariance, observe
    """

    def __init__(self, basis, coef_cov,
                 relative_precision=1.0E-15, **params):

        self.observed = False
        self.obs_mesh = None
        self.obs_V = None
        self.Uo = None
        self.obs_piv = None
        self.obs_len = None
        self.full_piv = None
        self.full_obs_mesh = None
        self.basiscov = True

        self.basis = ravel(basis)
        self.shape = self.get_shape_from_basis(basis)
        self.n = prod(self.shape)
        self.ndim = len(self.shape)

        self.relative_precision = relative_precision
        self.params = params

        if coef_cov.shape == self.shape:
            self.coef_cov = asmatrix(diag(coef_cov.ravel()))
        elif coef_cov.shape == self.shape * 2:
            self.coef_cov = asmatrix(coef_cov.reshape((self.n, self.n)))
        else:
            raise ValueError(
                "Covariance tensor's shape must be basis.shape or basis.shape*2 (using tuple multiplication).")

        # Cholesky factor the covariance matrix of the coefficients.
        U, m, piv = ichol_full(c=self.coef_cov, reltol=relative_precision)
        self.coef_U = asmatrix(U[:m, argsort(piv)])

        # Rank of the coefficient covariance.
        self.unobs_m = m
        self.m = m

        # Record the unobserved Cholesky factor of the coefficient covariance.
        self.unobs_coef_U = self.coef_U.copy()
        self.observed = False

    def get_shape_from_basis(self, basis):
        return shape(basis)

    def eval_basis(self, x, regularize=True):
        """
        basis_mat = C.eval_basis(x)

        Evaluates self's basis functions on x and returns them stacked
        in a matrix. basis_mat[i,j] gives basis function i evaluated at
        x[j,:].
        """
        if regularize:
            x = regularize_array(x)

        out = zeros((self.n, x.shape[0]), dtype=float, order='F')

        for i in xrange(self.n):
            out[i] = self.basis[i](x, **self.params)

        return out

    def cholesky(self, x, apply_pivot=True,
                 observed=True, nugget=None, regularize=True):
        __doc__ = Covariance.cholesky.__doc__

        if regularize:
            x = regularize_array(x)

        # The pivots are just 1:N, N being the shape of x.
        piv_return = arange(x.shape[1], dtype=int)

        # The rank of the Cholesky factor is just the rank of the
        # coefficient covariance matrix.
        if observed:
            coef_U = self.coef_U
            m = self.m
        else:
            coef_U = self.unobs_coef_U
            m = self.unobs_m

        # The Cholesky factor is the Cholesky factor of the basis times
        # the basis evaluated on x. This isn't triangular, but it does have
        # the property U.T*U = self(x,x).
        U_return = asmatrix(self.eval_basis(x, regularize=False))
        U_return = coef_U * U_return

        if apply_pivot:
            # Good for users.
            return U_return
        else:
            # Good for self.observe.
            return {'piv': piv_return, 'U': U_return}

    def continue_cholesky(self, x, x_old, chol_dict_old,
                          apply_pivot=True, observed=True, nugget=None):
        __doc__ = Covariance.continue_cholesky.__doc__

        # Stack the old and new x's.
        xtot = vstack((x_old, x))

        # Extract information about the Cholesky factor of self(x_old).
        U_old = chol_dict_old['U']
        m_old = U_old.shape[0]

        # The pivots will just be 1:N.
        piv_return = arange(xtot.shape[1], dtype=int)

        # The rank of the Cholesky factor is the rank of the Cholesky factor of the
        # coefficient covariance.
        if observed:
            coef_U = self.coef_U
            m = self.m
        else:
            coef_U = self.unobs_coef_U
            m = self.unobs_m

        # The number of old and new observations.
        N_old = x_old.shape[0]
        N_new = x.shape[0]

        # The Cholesky factor is the Cholesky factor of the coefficient covariance
        # times the basis function.
        U_new = asmatrix(self.eval_basis(x, regularize=False))
        U_new = coef_U * U_new

        U_return = hstack((U_old, U_new))

        if apply_pivot:
            # Good for users.
            return U_return
        else:
            # Good for self.observe.
            return {'piv': piv_return, 'U': U_return}

    def observe(self, obs_mesh, obs_V, output_type='o'):
        __doc__ = Covariance.observe.__doc__

        ndim = obs_mesh.shape[1]
        nobs = obs_mesh.shape[0]

        if self.ndim is not None:
            if not ndim == self.ndim:
                raise ValueError(
                    "Dimension of observation mesh is not equal to dimension of base mesh.")
        else:
            self.ndim = ndim

        # bases_o is the basis evaluated on the observation mesh.
        basis_o = asmatrix(self.eval_basis(obs_mesh, regularize=False))

        # chol(basis_o.T * coef_cov * basis_o)
        chol_inner = self.coef_U * basis_o

        if output_type == 's':
            C_eval = dot(chol_inner.T, chol_inner).copy('F')
            U_eval = linalg.cholesky(C_eval).T.copy('F')
            U = U_eval
            m = U_eval.shape[0]
            piv = arange(m)
        else:
            # chol(basis_o.T * coef_cov * basis_o + diag(obs_V)). Really should do this as low-rank update of covariance
            # of V.

            U, piv, m = ichol_basis(
                basis=chol_inner, nug=obs_V, reltol=self.relative_precision)

        U = asmatrix(U)
        piv_new = piv[:m]
        self.obs_piv = piv_new
        obs_mesh_new = obs_mesh[piv_new, :]

        self.Uo = U[:m, :m]

        # chol(basis_o.T * coef_cov * basis_o + diag(obs_V)) ^ -T * basis_o.T
        self.Uo_cov = trisolve(
            self.Uo,
            basis_o[:,
                    piv[:m]].T,
            uplo='U',
            transa='T')

        # chol(basis_o.T * coef_cov * basis_o + diag(obs_V)).T.I * basis_o.T *
        # coef_cov
        self.Uo_cov = self.Uo_cov * self.coef_cov

        # coef_cov = coef_cov - coef_cov * basis_o * (basis_o.T * coef_cov *
        # basis_o + diag(obs_V)).I * basis_o.T * coef_cov
        self.coef_cov = self.coef_cov - self.Uo_cov.T * self.Uo_cov

        # coef_U = chol(coef_cov)
        U, m, piv = ichol_full(c=self.coef_cov, reltol=self.relative_precision)
        U = asmatrix(U)
        self.coef_U = U[:m, argsort(piv)]
        self.m = m

        if output_type == 'o':
            return piv_new, obs_mesh_new

        if output_type == 's':
            return U_eval, C_eval, basis_o

        raise ValueError('Output type not recognized.')

    def __call__(self, x, y=None, observed=True,
                 regularize=True, return_Uo_Cxo=False):

        # Record the initial shape of x and regularize it.
        orig_shape = shape(x)
        if len(orig_shape) > 1:
            orig_shape = orig_shape[:-1]

        if regularize:
            x = regularize_array(x)

        ndimx = x.shape[-1]
        lenx = x.shape[0]

        # Get the correct version of the Cholesky factor of the coefficient
        # covariance.
        if observed:
            coef_U = self.coef_U
        else:
            coef_U = self.unobs_coef_U

        # Safety.
        if self.ndim is not None:
            if not self.ndim == ndimx:
                raise ValueError(
                    "The number of spatial dimensions of x does not match the number of spatial dimensions of the Covariance instance's base mesh.")

        # Evaluate the Cholesky factor of self's evaluation on x.
        # Will be observed or not depending on which version of coef_U
        # is used.
        basis_x_ = self.eval_basis(x, regularize=False)
        basis_x = coef_U * basis_x_

        # ==========================================================
        # = If only one argument is provided, return the diagonal: =
        # ==========================================================
        if y is None:
            # Diagonal calls done in Fortran for speed.
            V = basis_diag_call(basis_x)
            if return_Uo_Cxo:
                return V.reshape(orig_shape), basis_x_
            else:
                return V.reshape(orig_shape)

        # ===========================================================
        # = If the same argument is provided twice, save some work: =
        # ===========================================================
        if y is x:
            if return_Uo_Cxo:
                return basis_x.T * basis_x, basis_x_
            else:
                return basis_x

        # =========================================
        # = If y and x are different, do needful: =
        # =========================================
        else:

            # Regularize y and record its original shape.
            if regularize:
                y = regularize_array(y)

            ndimy = y.shape[-1]
            leny = y.shape[0]

            if not ndimx == ndimy:
                raise ValueError(
                    'The last dimension of x and y (the number of spatial dimensions) must be the same.')

            # Evaluate the Cholesky factor of self's evaluation on y.
            # Will be observed or not depending on which version of coef_U
            # is used.
            basis_y = self.eval_basis(y, regularize=False)
            basis_y = coef_U * basis_y

            return basis_x.T * basis_y

    # Helper methods for Mean instances.
    def _unobs_reg(self, M):
        # reg_mat = chol(self.basis_o.T * self.coef_cov * self.basis_o + diag(obs_V)).T.I * self.basis_o.T * self.coef_cov *
        # chol(self(obs_mesh_*, obs_mesh_*)).T.I * M.dev
        return self.Uo_cov.T * \
            asmatrix(trisolve(self.Uo, M.dev, uplo='U', transa='T')).T

    def _obs_reg(self, M, dev_new, m_old):
        # reg_mat = chol(self.basis_o.T * self.coef_cov * self.basis_o + diag(obs_V)).T.I * self.basis_o.T * self.coef_cov *
        # chol(self(obs_mesh_*, obs_mesh_*)).T.I * M.dev
        M.reg_mat = M.reg_mat + self.Uo_cov.T * asmatrix(
            trisolve(self.Uo, dev_new, uplo='U', transa='T')).T
        return M.reg_mat

    def _obs_eval(self, M, M_out, x, Uo_Cxo=None):
        basis_x = Uo_Cxo if Uo_Cxo is not None else self.eval_basis(
            x, regularize=False)
        M_out += asarray(dot(basis_x.T, M.reg_mat)).squeeze()
        return M_out

    def _mean_under_new(self, M, obs_mesh_new):
        if not M.observed:
            return asarray(M.eval_fun(obs_mesh_new, **M.params)).ravel()
        else:
            return M.__call__(obs_mesh_new, regularize=False)


class SeparableBasisCovariance(BasisCovariance):

    """
    C=SeparableBasisCovariance(basis, coef_cov, relative_precision, **params)

    Realizations generated using such covariances will be of the form:
        R(x) = \sum_{i[i],...,i[n]} basis[1][i[i]](x) * ... * basis[n][i[n]](x) * c[i[1],...,i[n]]
        c ~ N(mu, coef_cov)

    :Arguments:

        -   `basis`: An n-dimensional array of functions. Each should take an argument x
            of shape  (n,ndim), where n is any integer and ndim is the dimensionality of
            the space, or shape (n). In the latter case ndim should be assumed to be 1.

        -   `coef_cov`: An array of shape (i[1],...,i[n]) or (i[1],...,i[n])*2 (tuple multiplication). If
            shape=basis.shape, the coefficients are assumed to be independent. Otherwise
            coef_cov[i[1],...,i[n], j[1],...,j[n]] gives the prior covariance of coefficients
            i[1],...,i[n] and j[1],...,j[n].

        -   `params`: Parameters to be passed to basis.

        -   `relative_precision`: See documentation.


    :SeeAlso: Mean, Realization, Covariance, SeparableBasisCovariance, observe
    """

    def __init__(self, basis, coef_cov,
                 relative_precision=1.0E-15, **params):
        BasisCovariance.__init__(
            self,
            basis,
            coef_cov,
            relative_precision,
            **params)
        self.basis = basis
        self.n_per_dim = []
        for i in xrange(self.ndim):
            self.n_per_dim.append(len(self.basis[i]))

    def get_shape_from_basis(self, basis):
        return tuple([len(dim_basis) for dim_basis in basis])

    def eval_basis(self, x, regularize=True):
        """
        basis_mat = C.eval_basis(x)

        Evaluates self's basis functions on x and returns them stacked
        in a matrix. basis_mat[i,j] gives basis function i (formed by
        multiplying basis functions) evaluated at x[j,:].
        """
        # Make object of same shape as self.basis, fill in with evals of each individual basis factor.
        # Make object of same shape as diag(coef_cov), fill in with products of those evals.
        # Reshape and return.
        if regularize:
            x = regularize_array(x)

        out = zeros(self.shape + (x.shape[0],), dtype=float, order='F')

        # Evaluate the basis factors
        basis_factors = []
        for i in xrange(self.ndim):
            basis_factors.append([])
            for j in xrange(self.n_per_dim[i]):
                basis_factors[i].append(self.basis[i][j](x, **self.params))

        out = ones((self.n, x.shape[0]), dtype=float)
        out_reshaped = out.reshape(self.shape + (x.shape[0],))

        for ind in ndindex(self.shape):
            for dim in xrange(self.ndim):
                out_reshaped[ind] *= basis_factors[dim][ind[dim]]

        return out



import numpy as np
import imp
import pymc

from . import distances

from .distances import euclidean, aniso_geo_rad, paniso_geo_rad
from .distances import geographic as geo_rad


__all__ = ['euclidean', 'geo_rad', 'geo_deg', 'aniso_geo_rad', 'aniso_geo_deg', 'partition_aniso_geo_rad', 'partition_aniso_geo_deg']

euclidean.__name__ = 'euclidean'
euclidean.__doc__ = """
    euclidean(D,x,y,cmin=0,cmax=y.shape[0],symm=False)


    :Arguments:

        - `x and y` are arrays of points in Euclidean coordinates
          formatted as follows:

          [[x_{0,0} ... x_{0,ndim}],
           [x_{1,0} ... x_{1,ndim}],
           ...
           [x_{N,0} ... x_{N,ndim}]]

        - `symm` indicates whether x and y are references to
          the same array.

        - `cmin' and `cmax' indicate which columns to compute.
          These are used for multithreaded evaluation.


    Return value is a matrix D, where D[i,j] gives the Euclidean
    distance between the point x[i,:] and y[j,:].
"""

geo_rad.__name__ = 'geo_rad'
geo_rad.__doc__ = """
    geo_rad(D,x,y,cmin=0,cmax=y.shape[0],symm=False)


    :Arguments:

        - `x and y` are arrays of points in geographic coordinates
          formatted as follows:

          [[lon_0, lat_0],
           [lon_1, lat_1],
           ...
           [lon_N, lat_N]]

          Latitudes and longitudes should be in radians.

        - `cmin' and `cmax' indicate which columns to compute.
          These are used for multithreaded evaluation.

        - `symm` indicates whether x and y are references to
          the same array.

    Return value is a matrix D, where D[i,j] gives the great-circle
    distance between the point x[i,:] and y[j,:] on a sphere of unit
    radius.
"""

def geo_deg(D,x,y,cmin=0,cmax=-1,symm=False):
    geo_rad(D,x*np.pi/180., y*np.pi/180., cmin,cmax,symm)
    D *= 180./np.pi
geo_deg.__doc__ = geo_rad.__doc__.replace('radian', 'degree').replace('rad','deg')

aniso_geo_rad.extra_parameters = {'ecc': 'Eccentricity of level sets of distance', 'inc': 'Angle of inclination, in radians'}
aniso_geo_rad.__name__ = 'aniso_geo_rad'
aniso_geo_rad.__doc__ = """
    aniso_geo_rad(D,x,y,inc,ecc,cmin=0,cmax=y.shape[0],symm=False)


    :Arguments:

        - `x and y` are arrays of points in geographic coordinates
          formatted as follows:

          [[lon_0, lat_0],
           [lon_1, lat_1],
           ...
           [lon_N, lat_N]]

          Latitudes and longitudes should be in radians.

        - `inc` gives the eccentricity of the elliptical level sets of distance.

        - `ecc` gives the angle of inclination of the elliptical level sets of
          distance, in radians.

        - `cmin' and `cmax' indicate which columns to compute.
          These are used for multithreaded evaluation.

        - `symm` indicates whether x and y are references to
          the same array.

    Return value is a matrix D, where D[i,j] gives the great-circle
    distance between the point x[i,:] and y[j,:] on a sphere of unit
    radius.
"""

def aniso_geo_deg(D,x,y,inc,ecc,cmin=0,cmax=-1,symm=False):
    aniso_geo_rad(D,x*np.pi/180., y*np.pi/180., inc*np.pi/180., ecc, cmin, cmax, symm)
    D*=180./np.pi
aniso_geo_deg.__doc__ = geo_deg.__doc__.replace('radian', 'degree').replace('rad','deg')
aniso_geo_deg.extra_parameters = {'ecc': 'Eccentricity of level sets of distance', 'inc': 'Angle of inclination, in degrees'}

def partition_aniso_geo_rad(D,x,y,ctrs,scals,cmin=0,cmax=-1,symm=False):
    paniso_geo_rad(D,x,y,ctrs,scals,cmin,cmax,symm)
partition_aniso_geo_rad.extra_parameters = {'ctrs': 'Centers of angular bins, in radians','scals': 'Scales associated with each angular bin'}
partition_aniso_geo_rad.__doc__ = ""

def partition_aniso_geo_deg(D,x,y,ctrs,scals,cmin=0,cmax=-1,symm=False):
    paniso_geo_rad(D,x*np.pi/180., y*np.pi/180., ctrs*np.pi/180., scals, cmin,cmax,symm)
    D*=180./np.pi
partition_aniso_geo_deg.extra_parameters = {'ctrs': 'Centers of angular bins, in radians','scals': 'Scales associated with each angular bin'}
partition_aniso_geo_deg.__doc__ = ""



# Copyright (c) Anand Patil, 2007

import numpy as np
from . import wrapped_distances
import inspect
import imp
import pickle
from .isotropic_cov_funs import symmetrize, imul
from copy import copy
import sys,os
from pymc import get_threadpool_size, map_noreturn
import pymc

from pymc import six
xrange = six.moves.xrange

mod_search_path = [pymc.__path__[0]+'/gp/cov_funs', os.getcwd()] + sys.path


__all__ = ['covariance_wrapper', 'covariance_function_bundle']


def regularize_array(A):
    """
    Takes an np.ndarray as an input.


    - If the array is one-dimensional, it's assumed to be an array of input values.

    - If the array is more than one-dimensional, its last index is assumed to curse
      over spatial dimension.


    Either way, the return value is at least two dimensional. A.shape[-1] gives the
    number of spatial dimensions.
    """
    if not isinstance(A,np.ndarray):
        A = np.array(A, dtype=float)
    else:
        A = np.asarray(A, dtype=float)

    if len(A.shape) <= 1:
        return A.reshape(-1,1)

    elif A.shape[-1]>1:
        return A.reshape(-1, A.shape[-1])

    else:
        return A

def import_item(name):
    """
    Useful for importing nested modules such as pymc.gp.cov_funs.isotropic_cov_funs.

    Updated with code copied from IPython under a BSD license.
    """
    package = '.'.join(name.split('.')[0:-1])
    obj = name.split('.')[-1]

    if package:
        module = __import__(package,fromlist=[obj])
        return module.__dict__[obj]
    else:
        return __import__(obj)


class covariance_wrapper(object):
    """
    A wrapper for the Fortran covariance functions that
    removes the need for worrying about the common arguments
    amp and scale, and that separates the distance-finding
    from the covariance-evaluating for less code duplication
    and easier nonstationary generalizations.

    Parallelizes calls of sufficient size.
    """

    # pickle support
    def __getstate__(self):
        return (self.cov_fun_name, self.cov_fun_module.__name__, self.extra_cov_params,
                self.distance_fun_name, self.distance_fun_module.__name__)

    def __setstate__(self, state):
        self.__init__(*state)

    def __init__(self, cov_fun_name, cov_fun_module, extra_cov_params, distance_fun_name, distance_fun_module, with_x=False):

        self.cov_fun_name = cov_fun_name
        self.distance_fun_name = distance_fun_name

        try:
            cov_fun = import_item(cov_fun_module + "." + cov_fun_name)
        except ImportError:
            cov_fun_module = 'pymc.gp.cov_funs.' + cov_fun_module
            cov_fun = import_item(cov_fun_module + "." + cov_fun_name)

        try:
            distance_fun = import_item(distance_fun_module+"."+distance_fun_name)
        except ImportError:
            distance_fun_module = 'pymc.gp.cov_funs.' + distance_fun_module
            distance_fun = import_item(distance_fun_module+"."+distance_fun_name)

        self.cov_fun_module = sys.modules[cov_fun_module]
        self.cov_fun = cov_fun
        self.distance_fun_module = sys.modules[distance_fun_module]
        self.distance_fun = distance_fun
        self.extra_cov_params = extra_cov_params
        self.__doc__ = cov_fun_name + '.' + distance_fun.__name__+ covariance_wrapperdoc[0]
        self.with_x = with_x

        # Add covariance parameters to function signature
        for parameter in extra_cov_params:
            self.__doc__ += ', ' + parameter
        # Add distance parameters to function signature
        if hasattr(distance_fun,'extra_parameters'):
            self.extra_distance_params = distance_fun.extra_parameters
            for parameter in self.extra_distance_params:
                self.__doc__ += ', ' + parameter
        # Document covariance parameters
        self.__doc__ += covariance_wrapperdoc[1]
        if hasattr(cov_fun, 'extra_parameters'):
            for parameter in extra_cov_params:
                self.__doc__ += "\n\n    - " + parameter + ": " + extra_cov_params[parameter]
        # Document distance parameters.
        if hasattr(distance_fun,'extra_parameters'):
            for parameter in self.extra_distance_params:
                self.__doc__ += "\n\n    - " + parameter + ": " + self.extra_distance_params[parameter]

        self.__doc__ += "\n\nDistances are computed using "+distance_fun.__name__+":\n\n"+distance_fun.__doc__

    def __call__(self,x,y,amp=1.,scale=1.,symm=None,*args,**kwargs):

        if amp<0. or scale<0.:
            raise ValueError('The amp and scale parameters must be positive.')

        if symm is None:
            symm = (x is y)

        # Figure out how to divide job up between threads.
        nx = x.shape[0]
        ny = y.shape[0]
        n_threads = min(get_threadpool_size(), nx*ny // 10000)

        if n_threads > 1:
            if not symm:
                bounds = np.linspace(0,ny,n_threads+1)
            else:
                bounds = np.array(np.sqrt(np.linspace(0,ny*ny,n_threads+1)),dtype=int)

        # Split off the distance arguments
        distance_arg_dict = {}
        if hasattr(self.distance_fun, 'extra_parameters'):
            for key in self.extra_distance_params:
                if key in kwargs.keys():
                    distance_arg_dict[key] = kwargs.pop(key)

        # Allocate the matrix
        C = np.asmatrix(np.empty((nx,ny),dtype=float,order='F'))

        def targ(C,x,y, cmin, cmax,symm, d_kwargs=distance_arg_dict, c_args=args, c_kwargs=kwargs):
            # Compute distance for this bit
            self.distance_fun(C, x, y, cmin=cmin, cmax=cmax, symm=symm, **d_kwargs)
            imul(C, 1./scale, cmin=cmin, cmax=cmax, symm=symm)
            # Compute covariance for this bit
            if self.with_x:
                self.cov_fun(C,x,y,cmin=cmin, cmax=cmax,symm=symm,*c_args,**c_kwargs)
            else:
                self.cov_fun(C, cmin=cmin, cmax=cmax,symm=symm, *c_args, **c_kwargs)
            imul(C, amp*amp, cmin=cmin, cmax=cmax, symm=symm)

        if n_threads <= 1:
            targ(C,x,y,0,-1,symm)
        else:
            thread_args = [(C,x,y,bounds[i],bounds[i+1],symm) for i in xrange(n_threads)]
            map_noreturn(targ, thread_args)

        if symm:
            symmetrize(C)


        return C

class covariance_wrapper_with_diag(covariance_wrapper):
    def diag_call(self, x, amp=1., scale=1., *args, **kwds):
        return amp**2*np.ones(x.shape[0])

# Common verbiage in the covariance functions' docstrings
covariance_wrapperdoc = ["(x,y",""", amp=1., scale=1.)

A covariance function. Remember, broadcasting for covariance functions works
differently than for numpy universal functions. C(x,y) returns a matrix, and
C(x) returns a vector.


:Arguments:

    - `x, y`: arrays on which to evaluate the covariance function.

    - `amp`: The pointwise standard deviation of f.

    - `scale`: The factor by which to scale the distance between points.
             Large value implies long-range correlation."""]


class covariance_function_bundle(object):
    """
    B = covariance_function_bundle(cov_fun)

    A bundle of related covariance functions that use the stationary,
    isotropic covariance function cov_fun.

    Attributes:

        - `raw`: The raw covariance function, which overwrites a
          distance matrix with a covariance matrix.

        - `euclidean`: The covariance function wrapped to use
          Euclidean coordinates in R^n, with amp and scale arguments.

        - `geo_rad`: The covariance function wrapped to use
          geographic coordinates (latitude and longitude) on the
          surface of the sphere, with amp and scale arguments.

          Angles are assumed to be in radians. Radius of sphere is
          assumed to be 1, but you can effectively change the radius
          using the 'scale' argument.

        - `geo_deg`: Like geo_rad, but angles are in degrees.

        - `aniso_geo_rad`: Like geo_rad, but the distance function takes extra
          parameters controlling the eccentricity and angle of inclination of
          the elliptical level sets of distance.

        - `aniso_geo_deg`: Like aniso_geo_rad, but angles are in degrees.

        - `nonstationary`: Not implemented yet.

    Method:

        - `add_distance_metric(distance_fun)`: Takes a function that computes a
          distance matrix for points in some coordinate system and returns
          the covariance function wrapped to use that coordinate system.

    :Parameters:

        - `cov_fun` should overwrite distance matrices with covariance
          matrices in-place. In addition to the distance matrix, it should
          take an optional argument called 'symm' which indicates whether
          the output matrix will be symmetric.
    """

    def __init__(self, cov_fun_name, cov_fun_module, extra_cov_params, ampsq_is_diag=False, with_x=False):

        self.cov_fun_name = cov_fun_name
        self.cov_fun_module = cov_fun_module
        self.extra_cov_params = extra_cov_params
        self.ampsq_is_diag = ampsq_is_diag

        self.wrappers = []

        self.add_distance_metric('euclidean','wrapped_distances',with_x=with_x)
        self.add_distance_metric('geo_rad','wrapped_distances',with_x=with_x)
        self.add_distance_metric('geo_deg','wrapped_distances',with_x=with_x)
        self.add_distance_metric('aniso_geo_rad','wrapped_distances',with_x=with_x)
        self.add_distance_metric('aniso_geo_deg','wrapped_distances',with_x=with_x)
        self.add_distance_metric('partition_aniso_geo_deg','wrapped_distances',with_x=with_x)
        self.add_distance_metric('partition_aniso_geo_rad','wrapped_distances',with_x=with_x)

        self.raw = self.euclidean.cov_fun

    def add_distance_metric(self, distance_fun_name, distance_fun_module, with_x):
        """
        Takes a function that computes a distance matrix for
        points in some coordinate system and returns self's
        covariance function wrapped to use that distance function.


        Uses function apply_distance, which was used to produce
        self.euclidean and self.geographic and their docstrings.


        :Parameters:

            - `distance_fun`: Creates a distance matrix from two
              np.arrays of points, where the first index iterates
              over separate points and the second over coordinates.

              In addition to the arrays x and y, distance_fun should
              take an argument called symm which indicates whether
              x and y are the same array.


        :SeeAlso:
            - `apply_distance()`
        """

        if self.ampsq_is_diag:
            kls = covariance_wrapper_with_diag
        else:
            kls = covariance_wrapper
        new_fun = kls(self.cov_fun_name, self.cov_fun_module, self.extra_cov_params, distance_fun_name, distance_fun_module, with_x=with_x)
        self.wrappers.append(new_fun)
        # try:
        setattr(self, distance_fun_name, new_fun)
        # except:
        #     pass
        return new_fun
